# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import atexit
from optparse import OptionParser
import os
from subprocess import PIPE
from subprocess import Popen
import sys

from instrumental.api import Coverage
from instrumental.compat import execfile
from instrumental.recorder import ExecutionRecorder
from instrumental.reporting import ExecutionReport

parser = OptionParser(usage="instrumental [options] COMMAND ARG1 ARG2 ...")
parser.disable_interspersed_args()
parser.add_option('-r', '--report', dest='report',
                  action='store_true',
                  help='Print a detailed coverage report')
parser.add_option('-s', '--summary', dest='summary',
                  action='store_true',
                  help='Print a summary coverage report')
parser.add_option('-S', '--statements', dest='statements',
                  action='store_true',
                  help='Print a summary statement coverage report')
parser.add_option('-x', '--xml', dest='xml',
                  action='store_true',
                  help='Create a cobertura-compatible xml coverage report')
parser.add_option('--html', dest='html',
                  action='store_true',
                  help='Create an html coverage report')
parser.add_option('-a', '--all', dest='all',
                  action='store_true', default=False,
                  help='Show all constructs (not just those missing coverage')
parser.add_option('-t', '--target', dest='targets',
                  action='append', default=[],
                  help=('A Python regular expression; modules with names'
                        ' matching this regular expression will be'
                        ' instrumented and have their coverage reported'))
parser.add_option('-i', '--ignore', dest='ignores',
                  action='append', default=[],
                  help=('A Python regular expression; modules with names'
                        ' matching this regular expression will be'
                        ' ignored and not have their coverage reported'))
parser.add_option('--report-literals',
                  dest='report_conditions_with_literals',
                  action='store_true', default=False,
                  help=('Report conditions containing literals as though'
                        'they were reachable'))
parser.add_option('--ignore-assertions',
                  dest='instrument_assertions',
                  action='store_false', default=True,
                  help=('Do not instrument and report on the expressions'
                        ' present in assertions'))
parser.add_option('--use-metadata-cache',
                  dest='use_metadata_cache',
                  action='store_true', default=False,
                  help=('Use a metadata cache to (possibly) speed up'
                        ' execution of the target program'))
parser.add_option('--ignore-comparisons',
                  dest='instrument_comparisons',
                  action='store_false', default=True,
                  help=('Do not instrument comparison expressions'))

def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    
    opts, args = parser.parse_args(argv)
    
    if len(args) < 1:
        parser.print_help()
        sys.exit()
    
    if not opts.targets:
        sys.stdout.write("No targets specified. Use the '-t' option to specify packages to cover")
        sys.exit()
    
    coverage = Coverage(opts)
    coverage.start(opts.targets, opts.ignores)
    
    xml_filename = os.path.abspath('instrumental.xml')
    
    sourcefile = args[0]
    environment = {'__name__': '__main__',
                   '__file__': sourcefile,
                   }
    sys.argv = args[:]
    try:
        here = os.getcwd()
        execfile(sourcefile, environment)
    finally:
        if any([opts.summary,
                opts.report,
                opts.statements,
                opts.xml,
                opts.html]):
            coverage.stop()
            sys.stdout.write("\n")
            recorder = coverage.recorder
            report = ExecutionReport(here, recorder.metadata, opts)
            if opts.summary:
                sys.stdout.write(report.summary() + "\n")
            if opts.report:
                sys.stdout.write(report.report(opts.all) + "\n")
            if opts.statements:
                sys.stdout.write(report.statement_summary() + "\n")
            if opts.xml:
                report.write_xml_coverage_report(xml_filename)
            if opts.html:
                report.write_html_coverage_report()
            print



import inspect
import os
import shutil

from astkit import ast

from instrumental.instrument import CoverageAnnotator
from instrumental.recorder import ExecutionRecorder

def get_normalized_source(func):
    source = inspect.getsource(func)
    source_lines = source.splitlines(False)[1:]
    base_indentation = 0
    while source_lines[0][base_indentation] == ' ':
        base_indentation += 1
    normal_source =\
        "\n".join(line[base_indentation:] for line in source_lines)
    return normal_source

def load_module(func):
    normal_source = get_normalized_source(func)
    module = ast.parse(normal_source)
    return module, normal_source

class DummyConfig(object):
    instrument_assertions = True
    instrument_comparisons = True
    use_metadata_cache = False

class InstrumentationTestCase(object):
    
    def setup(self):
        # First clear out the recorder so that we'll create a new one
        ExecutionRecorder.reset()
        self.recorder = ExecutionRecorder.get()
        self.config = DummyConfig()
    
    def _instrument_module(self, module_func):
        from instrumental.metadata import MetadataGatheringVisitor
        from instrumental.pragmas import PragmaFinder
        module, source = load_module(module_func)
        pragmas = PragmaFinder().find_pragmas(source)
        metadata = MetadataGatheringVisitor.analyze(self.config,
                                                    module_func.__name__, 
                                                    'somemodule.py',
                                                    source, pragmas)
        self.recorder.add_metadata(metadata)
        transformer = CoverageAnnotator(self.config,
                                        module_func.__name__,
                                        self.recorder)
        inst_module = transformer.visit(module)
        # print renderer.render(inst_module)
        return inst_module
    
    def _assert_recorder_setup(self, module, starting_lineno=0):
        assert isinstance(module, ast.Module)
        
        assert isinstance(module.body[starting_lineno], ast.ImportFrom)
        assert module.body[starting_lineno].module == 'instrumental.recorder'
        assert isinstance(module.body[starting_lineno].names[0], ast.alias)
        assert module.body[starting_lineno].names[0].name == 'ExecutionRecorder'
        
        assert isinstance(module.body[starting_lineno+1], ast.Assign)
        assert isinstance(module.body[starting_lineno+1].targets[0], ast.Name)
        assert module.body[starting_lineno+1].targets[0].id == '_xxx_recorder_xxx_'
        assert isinstance(module.body[starting_lineno+1].value, ast.Call)
        assert isinstance(module.body[starting_lineno+1].value.func, ast.Attribute)
        assert isinstance(module.body[starting_lineno+1].value.func.value, ast.Name)
        assert module.body[starting_lineno+1].value.func.value.id == 'ExecutionRecorder'
        assert module.body[starting_lineno+1].value.func.attr == 'get'
        assert not module.body[starting_lineno+1].value.args
        assert not module.body[starting_lineno+1].value.keywords
        assert not module.body[starting_lineno+1].value.starargs
        assert not module.body[starting_lineno+1].value.kwargs
    
    def _assert_record_statement(self, statement, modname, lineno):
        assert isinstance(statement, ast.Expr), statement.__dict__
        assert isinstance(statement.value, ast.Call)
        assert isinstance(statement.value.func, ast.Attribute)
        assert isinstance(statement.value.func.value, ast.Name)
        assert statement.value.func.value.id == '_xxx_recorder_xxx_'
        assert statement.value.func.attr == 'record_statement'
        assert isinstance(statement.value.args[0], ast.Str)
        assert statement.value.args[0].s == modname
        assert isinstance(statement.value.args[1], ast.Num)
        assert statement.value.args[1].n == lineno

def setup():
    if os.path.exists('.instrumental.cache'):
        shutil.rmtree('.instrumental.cache')



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import os
import shutil
import sys

from mako.exceptions import text_error_template
from mako.lookup import TemplateLookup

class HTMLCoverageReport(object):
    
    def __init__(self, summary, sources):
        self.summary = summary
        self.sources = sources
    
    def write(self, directory):
        sys.stdout.write("writing html coverage report to %s\n" % directory)
        if os.path.isdir(directory):
            shutil.rmtree(directory)
        os.mkdir(directory)
        
        resource_directory = os.path.join(os.path.dirname(__file__), 
                                          'resources')
        shutil.copy(os.path.join(resource_directory, 'jquery-1.7.1.min.js'), 
                    directory)
        shutil.copy(os.path.join(resource_directory, 
                                 'jquery-ui-1.8.17.custom.min.js'), 
                    directory)
        shutil.copy(os.path.join(resource_directory, 'styles.css'), 
                    directory)
        shutil.copytree(os.path.join(resource_directory, 'ui-lightness'), 
                        os.path.join(directory, 'ui-lightness'))
        
        template_directory = os.path.join(os.path.dirname(__file__), 
                                          'templates')
        template_lookup = TemplateLookup(directories=[template_directory],
                                         output_encoding='utf-8')
        
        packages_template = template_lookup.get_template('/html/packages.html')
        packages_html = packages_template.render(summary=self.summary)
        packages_path = os.path.join(directory, 'packages.html')
        with open(packages_path, 'w') as f:
            f.write(packages_html)
        
        for package, package_summary in self.summary.packages.items():
            for module, module_summary in package_summary.modules.items():
                module_template = \
                    template_lookup.get_template('/html/module.html')
                conditions_by_line = {}
                for condition in module_summary.conditions:
                    conditions_by_col = \
                        conditions_by_line.setdefault(condition.lineno, {})
                    condition_list = conditions_by_col.setdefault(condition.node.col_offset, [])
                    condition_list.append(condition)
                try:
                    source = self.sources[module].decode('utf-8')
                    module_html = \
                        module_template.render(module=module_summary,
                                               source=source,
                                               conditions=conditions_by_line)
                except:
                    sys.stdout.write(text_error_template().render())
                    raise
                module_path = os.path.join(directory, module + '.html')
                with file(module_path, 'w') as f:
                    f.write(module_html)



import gc
import os
import sys
from xml.etree import ElementTree

ZERO = '0.000000'
ONE = '1.000000'

class DummyConfig(object):
    instrument_assertions = True
    instrument_comparisons = True
    use_metadata_cache = False
    report_conditions_with_literals = False

class TestXMLReport(object):
    
    def setUp(self):
        self.config = DummyConfig()
    
    def _run_test(self, *args, **kwargs):
        from instrumental import api
        from instrumental.recorder import ExecutionRecorder
        ExecutionRecorder.reset()
        c = api.Coverage(self.config)
        modname = 'instrumental.test.samples.robust'
        c.start([modname], [])
        if modname in sys.modules:
            robust = sys.modules[modname]
            reload(robust)
        else:
            from instrumental.test.samples import robust
        robust.test_func(*args, **kwargs)
        c.stop()
        return c.recorder
    
    def _verify_element(self, element, spec):
        assert element.tag == spec.pop('tag')
        child_specs = spec.pop('children', [])
        for attr, expected in spec.items():
            assert element.attrib[attr] == expected,(
                (attr, element.attrib[attr], expected))
        children = list(element)
        assert len(children) == len(child_specs)
        for child, child_spec in zip(children, child_specs):
            self._verify_element(child, child_spec)
    
    def test_full_coverage_constructs(self):
        from instrumental import constructs
        from instrumental.reporting import ExecutionReport
        
        recorder = self._run_test(True, True, False, True, False)
        metadata = recorder.metadata['instrumental.test.samples.robust']
        
        assert 10 == len(metadata.constructs), metadata.constructs.keys()
        
        decision = metadata.constructs['5.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 2 == decision.number_of_conditions(False)
        assert 1 == decision.conditions_missed(False)
        
        boolop = metadata.constructs['5.2']
        assert isinstance(boolop, constructs.LogicalAnd)
        assert 3 == boolop.number_of_conditions(False)
        assert 2 == len(boolop.conditions_missed(False))
        
        decision = metadata.constructs['8.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 2 == decision.number_of_conditions(False)
        assert 1 == decision.conditions_missed(False)
        
        boolop = metadata.constructs['8.2']
        assert isinstance(boolop, constructs.LogicalOr)
        assert 3 == boolop.number_of_conditions(False)
        assert 2 == len(boolop.conditions_missed(False))
        
        decision = metadata.constructs['11.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 2 == decision.number_of_conditions(False)
        assert 0 == decision.conditions_missed(False)
        
        compare = metadata.constructs['11.2']
        assert isinstance(compare, constructs.Comparison)
        assert 2 == compare.number_of_conditions(False)
        assert 0 == compare.conditions_missed(False)
        
        decision = metadata.constructs['14.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 2 == decision.number_of_conditions(False)
        assert 1 == decision.conditions_missed(False)
        
        decision = metadata.constructs['16.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 1 == decision.number_of_conditions(False)
        assert 0 == decision.conditions_missed(False)
        
        boolop = metadata.constructs['16.2']
        assert isinstance(boolop, constructs.LogicalOr)
        assert 2 == boolop.number_of_conditions(False)
        assert 1 == len(boolop.conditions_missed(False))
        
        boolop = metadata.constructs['16.3']
        assert isinstance(boolop, constructs.LogicalAnd)
        assert 2 == boolop.number_of_conditions(False)
        assert 1 == len(boolop.conditions_missed(False))

    def test_full_coverage_report(self):
        from instrumental.reporting import ExecutionReport
        
        recorder = self._run_test(True, True, False, True, False)
        report = ExecutionReport(os.getcwd(), recorder.metadata, self.config)
        xml_filename = 'test-xml-report.xml'
        report.write_xml_coverage_report(xml_filename)
        
        tree = ElementTree.parse(xml_filename)
        root = tree.getroot()
        
        lines_spec = {'tag': 'lines',
                      'children': [{'tag': 'line',
                                    'hits': '1',
                                    'line': '1',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '3',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '5',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '6',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '8',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '9',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '11',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '12',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '14',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '16',
                                    },
                                   ]}
        class_spec = {'tag': 'class',
                      'name': 'instrumental.test.samples.robust',
                      'filename': 'instrumental/test/samples/robust.py',
                      'condition-rate': '0.500000',
                      'branch-rate': '0.666667',
                      'line-rate': ONE,
                      'children': [{'tag': 'methods'}, lines_spec],
                      }
        root_children_spec = (
            {'tag': 'packages',
             'children': [{'tag': 'package',
                           'name': 'instrumental.test.samples',
                           'condition-rate': '0.500000',
                           'branch-rate': '0.666667',
                           'line-rate': ONE,
                           'children': [{'tag': 'classes',
                                         'children': [class_spec],
                                         }]
                           }]
             })
        spec = {'tag': 'coverage',
                'condition-rate': '0.500000',
                'branch-rate': '0.666667',
                'line-rate': ONE,
                'children': [root_children_spec]
                }
        self._verify_element(root, spec)
        os.remove(xml_filename)
    
    def test_partial_coverage_constructs(self):
        from instrumental import constructs
        from instrumental.reporting import ExecutionReport
        
        recorder = self._run_test(True, False, False, True, False, 0)
        metadata = recorder.metadata['instrumental.test.samples.robust']
        
        assert 10 == len(metadata.constructs), metadata.constructs.keys()
        
        decision = metadata.constructs['5.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 2 == decision.number_of_conditions(False)
        assert 1 == decision.conditions_missed(False)
        
        boolop = metadata.constructs['5.2']
        assert isinstance(boolop, constructs.LogicalAnd)
        assert 3 == boolop.number_of_conditions(False)
        assert 2 == len(boolop.conditions_missed(False))
        
        decision = metadata.constructs['8.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 2 == decision.number_of_conditions(False)
        assert 1 == decision.conditions_missed(False)
        
        boolop = metadata.constructs['8.2']
        assert isinstance(boolop, constructs.LogicalOr)
        assert 3 == boolop.number_of_conditions(False)
        assert 2 == len(boolop.conditions_missed(False))
        
        decision = metadata.constructs['11.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 2 == decision.number_of_conditions(False)
        assert 1 == decision.conditions_missed(False)
        
        compare = metadata.constructs['11.2']
        assert isinstance(compare, constructs.Comparison)
        assert 2 == compare.number_of_conditions(False)
        assert 1 == compare.conditions_missed(False)
        
        decision = metadata.constructs['14.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 2 == decision.number_of_conditions(False)
        assert 1 == decision.conditions_missed(False)
        
        decision = metadata.constructs['16.1']
        assert isinstance(decision, constructs.BooleanDecision)
        assert 1 == decision.number_of_conditions(False)
        assert 0 == decision.conditions_missed(False)
        
        boolop = metadata.constructs['16.2']
        assert isinstance(boolop, constructs.LogicalOr)
        assert 2 == boolop.number_of_conditions(False)
        assert 1 == len(boolop.conditions_missed(False))
        
        boolop = metadata.constructs['16.3']
        assert isinstance(boolop, constructs.LogicalAnd)
        assert 2 == boolop.number_of_conditions(False)
        assert 1 == len(boolop.conditions_missed(False))

    def test_partial_coverage_report(self):
        from instrumental.reporting import ExecutionReport
        
        recorder = self._run_test(True, False, False, True, False, 0)
        
        report = ExecutionReport(os.getcwd(), recorder.metadata, self.config)
        xml_filename = 'test-xml-report.xml'
        report.write_xml_coverage_report(xml_filename)
        
        tree = ElementTree.parse(xml_filename)
        root = tree.getroot()
        
        lines_spec = {'tag': 'lines',
                      'children': [{'tag': 'line',
                                    'hits': '1',
                                    'line': '1',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '3',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '5',
                                    },
                                   {'tag': 'line',
                                    'hits': '0',
                                    'line': '6',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '8',
                                    },
                                   {'tag': 'line',
                                    'hits': '0',
                                    'line': '9',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '11',
                                    },
                                   {'tag': 'line',
                                    'hits': '0',
                                    'line': '12',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '14',
                                    },
                                   {'tag': 'line',
                                    'hits': '1',
                                    'line': '16',
                                    },
                                   ]}
        class_spec = {'tag': 'class',
                      'name': 'instrumental.test.samples.robust',
                      'filename': 'instrumental/test/samples/robust.py',
                      'condition-rate': '0.416667',
                      'branch-rate': '0.555556',
                      'line-rate': '0.700000',
                      'children': [{'tag': 'methods'}, lines_spec],
                      }
        root_children_spec = (
            {'tag': 'packages',
             'children': [{'tag': 'package',
                           'name': 'instrumental.test.samples',
                           'condition-rate': '0.416667',
                           'branch-rate': '0.555556',
                           'line-rate': '0.700000',
                           'children': [{'tag': 'classes',
                                         'children': [class_spec],
                                         }]
                           }]
             })
        spec = {'tag': 'coverage',
                'condition-rate': '0.416667',
                'branch-rate': '0.555556',
                'line-rate': '0.700000',
                'children': [root_children_spec]
                }
                
        self._verify_element(root, spec)
        os.remove(xml_filename)



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import imp
import os
import re
import sys

from astkit import ast

from instrumental.compat import exec_f

_imp_load_module = imp.load_module
def monkeypatch_imp(targets, ignores, visitor_factory):
    imp.load_module = load_module_factory(targets, ignores, visitor_factory)

def unmonkeypatch_imp():
    imp.load_module = _imp_load_module

def load_module_factory(targets, ignores, visitor_factory):
    def load_module(name, fh, pathname, description):
        if ((not any([re.match(target, name) for target in targets]))
            or
            (any([re.match(ignore, name) for ignore in ignores]))):
            return _imp_load_module(name, fh, pathname, description)
        else:
            suffix, mode, type = description
            ispkg = type == imp.PKG_DIRECTORY
            if ispkg:
                source = open(os.path.join(pathname, '__init__.py'), 'r').read()
            else:
                source = fh.read()
            visitor = visitor_factory.create(name, source)
            code_tree = ast.parse(source)
            new_code_tree = visitor.visit(code_tree)
            code = compile(new_code_tree, pathname, 'exec')
            mod = sys.modules.setdefault(name, imp.new_module(name))
            if ispkg:
                mod.__file__ = os.path.join(pathname, '__init__.py')
                mod.__path__ = [pathname]
            else:
                mod.__file__ = pathname
            mod.__loader__ = None
            exec_f(code, mod.__dict__)
            return mod
    return load_module



import ast

from astkit.render import SourceCodeRenderer

from instrumental.constructs import BooleanDecision

class TestBooleanDecision(object):
    
    def setup(self):
        self.node = ast.Compare(left=ast.Name(id='x'),
                                ops=[ast.Eq()],
                                comparators=[ast.Num(n=4)],
                                lineno=5, col_offset=23)
        self.modulename = 'somepackage.somemodule'
        self.label = '5.1'

    def _makeOne(self):
        return BooleanDecision(self.modulename, self.label, self.node, {})
    
    def test_a_new_one(self):
        decision = self._makeOne()
        assert self.modulename == decision.modulename
        assert 5 == decision.lineno
        assert "(x == 4)" == decision.source
        assert {True: set(),
                False: set()} == decision.conditions
    
    def test_record_True(self):
        decision = self._makeOne()
        decision.record(True, 'X')
        assert decision.conditions[True] == set('X')
        assert not decision.conditions[False]
    
    def test_record_False(self):
        decision = self._makeOne()
        decision.record(False, 'X')
        assert not decision.conditions[True]
        assert decision.conditions[False] == set('X')
    
    def test_number_of_conditions(self):
        decision = self._makeOne()
        assert 2 == decision.number_of_conditions(False)
    
    def test_number_of_conditions_hit(self):
        decision = self._makeOne()
        decision.record(False, 'X')
        assert 1 == decision.number_of_conditions_hit()
    
    def test_conditions_missed(self):
        decision = self._makeOne()
        assert 2 == decision.conditions_missed(False)
    
    def _make_expected_result(self, decision, *conditions_hit):
        node_source = SourceCodeRenderer.render(self.node)
        result_lines = []
        result_lines.append("Decision -> %s:%s < %s >" % (self.modulename,
                                                          self.label,
                                                          node_source)
                            )
        result_lines.append("")
        for condition in ("T", "F"):
            tag = 'X' if condition in conditions_hit else ''
            result_lines.append("%s ==> %s" % (condition, tag))
        return "\n".join(result_lines)
    
    def test_result_not_hit(self):
        decision = self._makeOne()
        expected = self._make_expected_result(decision)
        assert expected == decision.result(),\
            (expected, decision.result())



from astkit import ast

from instrumental.recorder import ExecutionRecorder

class KnownValue(object):
    pass

class TestRecorder(object):
    def setup(self):
        # Reset recorder
        ExecutionRecorder.reset()
    
    def test_construct_with_literal(self):
        recorder = ExecutionRecorder.get()
        node = ast.BoolOp(op=ast.Or(),
                          values=[ast.Name(id="foo"),
                                  ast.Str(s='""')],
                          lineno=1,
                          col_offset=0)
        recorder.add_BoolOp('somemodule', '1.1', node, [], None)
    
    def test_add_a_non_BoolOp(self):
        recorder = ExecutionRecorder.get()
        node = ast.BoolOp(op=4,
                          values=[ast.Name(id="foo"),
                                  ast.Str(s='""')],
                          lineno=1,
                          col_offset=0)
        try:
            recorder.add_BoolOp('somemodule', node, [], None)
        except TypeError as exc:
            assert "BoolOp" in str(exc), exc




from copy import deepcopy
import fnmatch
import itertools
import os
import pickle
import re
import sys
import time

from astkit import ast

from instrumental import constructs
from instrumental import util
from instrumental.pragmas import PragmaFinder
from instrumental.pragmas import PragmaNoCover

def has_docstring(defn):
    return ast.get_docstring(defn) is not None

def gather_metadata(config, recorder, targets, ignores):
    finder = SourceFinder(sys.path)
    if config.use_metadata_cache:
        metadata_cache = FileBackedMetadataCache()
    else:
        metadata_cache = DummyMetadataCache()
    for target in targets:
        for source_spec in finder.find(target, ignores):
            filepath, modulename = source_spec
            metadata = metadata_cache.fetch(filepath)
            if not metadata:
                source = open(filepath, "r").read()
                pragmas = PragmaFinder().find_pragmas(source)
                metadata = MetadataGatheringVisitor.analyze(config,
                                                            modulename,
                                                            filepath,
                                                            source,
                                                            pragmas)
                metadata_cache.store(filepath, metadata)
            recorder.add_metadata(metadata)

class ModuleMetadata(object):
    
    def __init__(self, modulename, filepath, source, pragmas):
        self.modulename = modulename
        self.filepath = filepath
        self.source = source
        self.lines = {}
        self.constructs = {}
        self.pragmas = pragmas
    
    def is_package(self):
        return self.filepath.endswith('__init__.py')
    
    def next_label(self, lineno):
        i = 1
        while ('%s.%s' % (lineno, i)) in self.constructs:
            i += 1
        return '%s.%s' % (lineno, i)

class BooleanEvaluator(ast.NodeVisitor):
    
    @classmethod
    def evaluate(cls, node):
        evaluator = cls()
        evaluator.visit(node)
        return evaluator._results
    
    def __init__(self):
        self._results = set([True, False])
    
    def _and(self, left, right):
        return left and right
    
    def _or(self, left, right):
        return left or right
    
    def _booleval(self, op, left, right):
        if isinstance(op, ast.And):
            return self._and(left, right)
        elif isinstance(op, ast.Or):
            return self._or(left, right)
        else: # pragma: no cover
            raise TypeError('Invalid operation: %s', op)
    
    def visit_BoolOp(self, node):
        for subexpr in node.values:
            subresults = BooleanEvaluator.evaluate(subexpr)
            pairs = itertools.product(self._results, subresults)
            self._results = set(self._booleval(node.op, result, subresult)
                                for result, subresult in pairs)
    
    def visit_IfExp(self, node):
        test_results = BooleanEvaluator.evaluate(node.test)
        body_results = BooleanEvaluator.evaluate(node.body)
        orelse_results = BooleanEvaluator.evaluate(node.orelse)
        
        if len(test_results) == 1:
            test_result = list(test_results)[0]
            if test_result:
                self._results = body_results
            else:
                self._results = orelse_results
        else:
            self._results = body_results | orelse_results
    
    def visit_Name(self, node):
        try:
            self._results = set([bool(ast.literal_eval(node))])
        except ValueError:
            pass
    
    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.Not):
            results = BooleanEvaluator.evaluate(node.operand)
            self._results = set(not result for result in results)
        else:
            self.visit_indeterminate(node)
    
    def visit_literal(self, node):
        self._results = set([bool(ast.literal_eval(node))])
    visit_Num = visit_literal
    visit_Str = visit_literal
    
    visit_Dict = visit_literal
    visit_List = visit_literal
    visit_Set = visit_literal
    visit_Tuple = visit_literal

    def visit_True(self, node):
        self._results = set([True])
    visit_GeneratorExp = visit_True
    visit_Lambda = visit_True
    
    def visit_indeterminate(self, node):
        pass
    visit_Attribute = visit_indeterminate
    visit_BinOp = visit_indeterminate
    visit_Call = visit_indeterminate
    visit_Compare = visit_indeterminate
    visit_DictComp = visit_indeterminate
    visit_ListComp = visit_indeterminate
    visit_Repr = visit_indeterminate
    visit_SetComp = visit_indeterminate
    visit_Subscript = visit_indeterminate
    visit_Yield = visit_indeterminate
    
class MetadataGatheringVisitor(ast.NodeVisitor):
    
    @classmethod
    def analyze(cls, config, modulename, filepath, source, pragmas):
        module_ast = ast.parse(source)
        metadata = ModuleMetadata(modulename, filepath, source, pragmas)
        visitor = cls(config, metadata, pragmas)
        visitor.visit(module_ast)
        return visitor.metadata
    
    def __init__(self, config, metadata, pragmas):
        self.config = config
        self.metadata = metadata
        self.modifiers = []
        self.gather = True
        self._context = []
    
    def _has_pragma(self, pragma, lineno):
        return any(isinstance(p, pragma) for p in self.metadata.pragmas[lineno])
    
    def generic_visit(self, node):
        if isinstance(node, ast.stmt):
            if self._has_pragma(PragmaNoCover, node.lineno):
                self.modifiers.append(PragmaNoCover)
            if PragmaNoCover not in self.modifiers:
                self.metadata.lines[node.lineno] = False
            if isinstance(node, ast.ClassDef) or isinstance(node, ast.FunctionDef):
                docstring = None
                if has_docstring(node):
                    docstring = node.body.pop(0)
        super(MetadataGatheringVisitor, self).generic_visit(node)
        if isinstance(node, ast.stmt):
            if isinstance(node, ast.ClassDef) or isinstance(node, ast.FunctionDef):
                if docstring:
                    node.body.insert(0, docstring)
            if self._has_pragma(PragmaNoCover, node.lineno):
                self.modifiers.pop(-1)

    def _make_boolop_construct(self, label, node):
        if isinstance(node.op, ast.And):
            klass = constructs.LogicalAnd
        elif isinstance(node.op, ast.Or):
            klass = constructs.LogicalOr
        pragmas = self.metadata.pragmas.get(node.lineno, [])
        construct = klass(self.metadata.modulename, label, node, pragmas)
        return construct
    
    def _make_decision(self, label, node):
        pragmas = self.metadata.pragmas.get(node.lineno, [])
        construct = constructs.BooleanDecision(self.metadata.modulename,
                                               label, node, pragmas)
        possible_results = BooleanEvaluator.evaluate(node)
        if True not in possible_results:
            construct.set_unreachable(True)
        if False not in possible_results:
            construct.set_unreachable(False)
        return construct
    
    def _make_comparison(self, label, node):
        pragmas = self.metadata.pragmas.get(node.lineno, [])
        construct = constructs.Comparison(self.metadata.modulename, 
                                          label, node, pragmas)
        return construct
    
    def visit_Module(self, module):
        if has_docstring(module):
            module.body.pop(0)
        self.generic_visit(module)
    
    def visit_Assert(self, assert_):
        if self.config.instrument_assertions:
            if isinstance(assert_.test, ast.BoolOp):
                label = self.metadata.next_label(assert_.lineno)
                construct = self._make_decision(label, assert_.test)
                self.metadata.constructs[label] = construct
                self._context.append(construct)
            self.generic_visit(assert_)
            if isinstance(assert_.test, ast.BoolOp):
                self._context.pop()
    
    def visit_assignment(self, assign):
        if isinstance(assign.value, ast.BoolOp):
            label = self.metadata.next_label(assign.lineno)
            construct = self._make_decision(label, assign.value)
            self.metadata.constructs[label] = construct
            self._context.append(construct)
        self.generic_visit(assign)
        if isinstance(assign.value, ast.BoolOp):
            self._context.pop()
    visit_Assign = visit_AugAssign = visit_assignment
    
    def visit_BoolOp(self, boolop):
        label = self.metadata.next_label(boolop.lineno)
        construct = self._make_boolop_construct(label, boolop)
        self.metadata.constructs[label] = construct
        self._context.append(construct)
        self.generic_visit(boolop)
        self._context.pop()
    
    def visit_Compare(self, compare):
        if self.config.instrument_comparisons:
            label = self.metadata.next_label(compare.lineno)
            construct = self._make_comparison(label, compare)
            self.metadata.constructs[label] = construct
            self._context.append(construct)
        self.generic_visit(compare)
        if self.config.instrument_comparisons:
            self._context.pop()
    
    def visit_If(self, if_):
        if self._has_pragma(PragmaNoCover, if_.lineno):
            self.modifiers.append(PragmaNoCover)
        if PragmaNoCover not in self.modifiers:
            self.metadata.lines[if_.lineno] = False
            label = self.metadata.next_label(if_.lineno)
            construct = self._make_decision(label, if_.test)
            self.metadata.constructs[str(label)] = construct
            self._context.append(construct)
            self.generic_visit(if_)
            self._context.pop()
        if self._has_pragma(PragmaNoCover, if_.lineno):
            self.modifiers.pop(-1)
    
    def visit_IfExp(self, ifexp):
        if self.gather:
            label = self.metadata.next_label(ifexp.lineno)
            construct = self._make_decision(label, ifexp.test)
            self.metadata.constructs[str(label)] = construct
            self._context.append(construct)
        self.generic_visit(ifexp)
        if self.gather:
            self._context.pop()
    
    def visit_While(self, while_):
        if self._has_pragma(PragmaNoCover, while_.lineno):
            self.modifiers.append(PragmaNoCover)
        if PragmaNoCover not in self.modifiers:
            self.metadata.lines[while_.lineno] = False
            label = self.metadata.next_label(while_.lineno)
            construct = self._make_decision(label, while_.test)
            self.metadata.constructs[str(label)] = construct
            self._context.append(construct)
            self.generic_visit(while_)
            self._context.pop()
        if self._has_pragma(PragmaNoCover, while_.lineno):
            self.modifiers.pop(-1)

class SourceFinder(object):
    """ Searches a given path for source files that meet criteria """
    
    def __init__(self, path):
        """ path is a list of paths that can contain source files """
        self.path = path
        if '.' not in self.path:
            self.path.append('.')
    
    def find(self, target, ignores):
        """ Find source files that look like `target` but not `ignores` """
        found = False
        for path in self.path:
            # We can only search directories
            if not os.path.isdir(path):
                continue
            
            for filepath in self._find_target(path, target, ignores):
                found = True
                modulename = filepath[len(path)+1:-3].replace(os.path.sep, '.')
                if modulename.endswith('__init__'):
                    modulename = modulename[:-9]
                if any(modulename.startswith(ignore) for ignore in ignores):
                    continue
                yield (filepath, modulename)
            if found:
                break
    
    def _is_python_source(self, filename):
        return filename == '__init__.py' or filename.endswith('.py')
    
    def _is_package_directory(self, dirname):
        return (os.path.isdir(dirname) and
                os.path.exists(os.path.join(dirname, '__init__.py')))
    
    def _find_target(self, path, target, ignores):
        if '.' in target:
            prefix, suffix = target.split('.', 1)
        else:
            prefix, suffix = target, '*'
        filenames = fnmatch.filter(sorted(os.listdir(path)), prefix)
        filenames += fnmatch.filter(sorted(os.listdir(path)), prefix + '.py')
        for filename in filenames:
            filepath = os.path.join(path, filename)
            if self._is_python_source(filename):
                yield filepath
            elif self._is_package_directory(filepath):
                for filepath in self._find_target(filepath, suffix, ignores):
                    yield filepath


class BaseMetadataCache(object):
    
    def initialize(self):
        self._init_storage()
    
    def store(self, filepath, meta):
        now = util.now()
        record = {'timestamp': now,
                  'metadata': meta}
        self._store(filepath, record)
    
    def fetch(self, filepath):
        file_mtime = os.stat(filepath).st_mtime
        cached_record = self._fetch(filepath)
        if not cached_record:
            return
        timestamp = time.mktime(cached_record['timestamp'].timetuple())
        if file_mtime < timestamp:
            return cached_record['metadata']

class DummyMetadataCache(BaseMetadataCache):
    
    def _init_storage(self):
        pass
    
    def _store(self, filepath, record):
        pass
    
    def _fetch(self, filepath):
        return None

class FileBackedMetadataCache(BaseMetadataCache):
    
    def __init__(self):
        super(FileBackedMetadataCache, self).__init__()
        self._working_directory = os.path.join(os.getcwd(), '.instrumental.cache')
        
    def _init_storage(self):
        if not os.path.exists(self._working_directory):
            os.mkdir(self._working_directory)
    
    def _store(self, filepath, record):
        if filepath.startswith('/'):
            filepath = filepath[1:]
        cache_file_path = os.path.join(self._working_directory, filepath)
        if not os.path.exists(os.path.dirname(cache_file_path)):
            os.makedirs(os.path.dirname(cache_file_path))
        with open(cache_file_path, 'wb') as cache_file:
            pickle.dump(record, cache_file)
    
    def _fetch(self, filepath):
        if filepath.startswith('/'):
            filepath = filepath[1:]
        cache_file_path = os.path.join(self._working_directory, filepath)
        if not os.path.exists(cache_file_path):
            return None
        with open(cache_file_path, 'rb') as cache_file:
            record = pickle.load(cache_file)
        return record


if __name__ == '__main__': # pragma: no cover

    target = sys.argv[1]
    ignores = [sys.argv[2]] if len(sys.argv) > 2 else []
    finder = SourceFinder(sys.path)
    for source_spec in finder.find(target, ignores):
        filepath, modulename = source_spec
        print filepath, modulename



import sys

from instrumental.importer import ImportHook
from instrumental.instrument import AnnotatorFactory
from instrumental.metadata import gather_metadata
from instrumental.monkey import monkeypatch_imp
from instrumental.monkey import unmonkeypatch_imp
from instrumental.recorder import ExecutionRecorder

class Coverage(object):
    
    def __init__(self, config):
        self._config = config
        self._import_hooks = []
    
    @property
    def recorder(self):
        return ExecutionRecorder.get()
    
    def start(self, targets, ignores):
        gather_metadata(self._config, self.recorder, targets, ignores)
        annotator_factory = AnnotatorFactory(self._config, self.recorder)
        monkeypatch_imp(targets, ignores, annotator_factory)
        for target in targets:
            hook = ImportHook(target, ignores, annotator_factory)
            self._import_hooks.append(hook)
            sys.meta_path.append(hook)
        self.recorder.start()
        
    def stop(self):
        self.recorder.stop()
        for hook in self._import_hooks:
            sys.meta_path.remove(hook)
        unmonkeypatch_imp()
    
    def start_context(self, label):
        self.recorder.tag = label
    
    def stop_context(self):
        self.recorder.tag = None



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import os

from instrumental.constructs import LogicalBoolean
from instrumental.htmlreport import HTMLCoverageReport
from instrumental.summary import ExecutionSummary
from instrumental.xmlreport import XMLCoverageReport

class ExecutionReport(object):
    
    def __init__(self, working_directory, metadata, options):
        self.working_directory = working_directory
        self.metadata = metadata
        self.options = options
    
    def report(self, showall=False):
        lines = []
        lines.append("")
        lines.append("===============================================")
        lines.append("Instrumental Condition/Decision Coverage Report")
        lines.append("===============================================")
        lines.append("")
        no_missed_conditions = True
        def _key_func(pair):
            label, _ = pair
            lineno, index = label.split('.')
            return (int(lineno), int(index))
        for modulename, metadata in sorted(self.metadata.items()):
            for label, construct in sorted(metadata.constructs.items(),
                                           key=_key_func):
                if showall or construct.conditions_missed(self.options.report_conditions_with_literals):
                    no_missed_conditions = False
                    lines.append(construct.result())
                    lines.append("")
        if no_missed_conditions:
            lines.append(" Full coverage. Good job!")
            lines.append("")
        return "\n".join(lines)
    
    def summary(self):
        lines = []
        lines.append("")
        lines.append("================================================")
        lines.append("Instrumental Condition/Decision Coverage Summary")
        lines.append("================================================")
        lines.append("")
        for modulename, metadata in sorted(self.metadata.items()):
            total_conditions = sum(construct.number_of_conditions(self.options.report_conditions_with_literals)
                                   for construct in metadata.constructs.values())
            hit_conditions = sum(construct.number_of_conditions_hit()
                                 for construct in metadata.constructs.values())
            if not total_conditions:
                lines.append('%s: 0/0 hit (---)' % modulename)
            else:
                lines.append('%s: %s/%s hit (%.0f%%)' %\
                                 (modulename, hit_conditions, total_conditions,
                                  hit_conditions/float(total_conditions) * 100))
        return '\n'.join(lines)
    
    def statement_summary(self):
        outlines = ["=======================================", 
                    "Instrumental Statement Coverage Summary", 
                    "=======================================",
                    "",
                    ]
        
        formatter = StatementCoverageFormatter()
        statements = dict((modulename, self.metadata[modulename].lines)
                          for modulename in self.metadata)
        return "\n".join(outlines + [formatter.format(statements)])

    def write_xml_coverage_report(self, filename):
        xml_report = XMLCoverageReport(self.working_directory, self.metadata, self.options)
        xml_report.write(filename)
    
    def write_html_coverage_report(self, directory='instrumental-result-html'):
        conditions = {}
        statements = {}
        sources = {}
        for modulename in self.metadata:
            conditions[modulename] = self.metadata[modulename].constructs
            statements[modulename] = self.metadata[modulename].lines
            sources[modulename] = self.metadata[modulename].source
        summary = ExecutionSummary(conditions, statements, self.options)
        html_report = HTMLCoverageReport(summary, sources)
        html_report.write(os.path.join(self.working_directory, directory))
    
class Chunk(object):
    
    def __init__(self, start):
        self.start = self.finish = start
    
    def extend(self):
        self.finish += 1
    
    def __str__(self):
        if self.start == self.finish:
            return str(self.start)
        elif self.start + 1 == self.finish:
            return "%s,%s" % (self.start, self.finish)
        else:
            return "%s-%s" % (self.start, self.finish)

class StatementCoverageFormatter(object):
    
    def _collapse_sequence(self, seq):
            
        chunks = []
        last_chunk = None
        for line in seq:
            if (not chunks) or (line != (chunks[-1].finish + 1)):
                chunks.append(Chunk(line))
            else:
                chunks[-1].extend()
        
        return ",".join([str(chunk) for chunk in chunks])
    
    def format(self, statements):
        header = self._make_header(statements)
        separator = self._make_separator(statements)
        lines = self._make_lines(statements)
        summary = self._make_summary(statements)
        
        return "\n".join([header, separator] + lines + [separator, summary])
    
    def _make_header(self, statements):
        longest_name_length = max(len(modulename) for modulename in statements)
        
        header = "Name" + (" " * longest_name_length)
        header = "".join([header, "Stmts"])
        header = "".join([header, (" " * 3) + "Miss"])
        header = "".join([header, (" " * 2) + "Cover"])
        header = "".join([header, (" " * 3) + "Missing"])
        
        return header
    
    def _make_separator(self, statements):
        header = self._make_header(statements)
        return "-" * len(header)
    
    def _make_line(self, modulename, lines, column_width):
        missing_lines = [line for line in sorted(lines) if not lines[line]]
        if lines:
            cover_pct = "%s%%" % int(100 * (len(lines) - len(missing_lines)) / float(len(lines)))
        else:
            cover_pct = "100%"
        
        line = modulename.ljust(column_width)
        line = "".join([line, str(len(lines)).rjust(5)])
        line = "".join([line, str(len(missing_lines)).rjust(7)])
        line = "".join([line, cover_pct.rjust(7)])
        line = "".join([line, 
                        (" " * 3), 
                        self._collapse_sequence(missing_lines)])
        
        return line
    
    def _make_lines(self, statements):
        longest_name_length = max(len(modulename) for modulename in statements)
        
        return [self._make_line(modulename, lines, longest_name_length + 4)
                for modulename, lines 
                in sorted(statements.items(),
                          key=lambda pair: pair[0])]
    
    def _make_summary(self, statements):
        longest_name_length = max(len(modulename) for modulename in statements)
        
        total_lines = 0
        missed_lines = 0
        for lines in statements.values():
            total_lines += len(lines)
            missed_lines += len([a_line for a_line in lines
                                 if not lines[a_line]])
        covered_pct = "%.0f%%" % int(100 * float(total_lines - missed_lines) / total_lines)
        
        summary = "TOTAL".ljust(longest_name_length + 4)
        summary = "".join([summary, str(total_lines).rjust(5)])
        summary = "".join([summary, str(missed_lines).rjust(7)])
        summary = "".join([summary, covered_pct.rjust(7)])
        
        return summary



import sys

from astkit.render import SourceCodeRenderer as renderer

from instrumental.compat import exec_f
from instrumental.instrument import CoverageAnnotator
from instrumental.recorder import ExecutionRecorder
from instrumental.test import DummyConfig
from instrumental.test import load_module

class TestInstrumentation(object):
    
    def setup(self):
        # First clear out the recorder so that we'll create a new one
        ExecutionRecorder.reset()
        self.recorder = ExecutionRecorder.get()
        self.recorder.start()
    
    def teardown(self):
        self.recorder.stop()
    
    def _load_and_compile_module(self, module_func):
        module, source = load_module(module_func)
        from instrumental.pragmas import PragmaFinder
        pragmas = PragmaFinder().find_pragmas(source)
        from instrumental.metadata import MetadataGatheringVisitor
        config = DummyConfig()
        self.recorder.add_metadata(MetadataGatheringVisitor.analyze(config,
                                                                    module_func.__name__,
                                                                    'somemodule.py',
                                                                    source, 
                                                                    pragmas))
        # self.recorder.add_source(module_func.__name__, source)
        transformer = CoverageAnnotator(config,
                                        module_func.__name__,
                                        self.recorder)
        inst_module = transformer.visit(module)
        sys.stdout.write(renderer.render(inst_module) + "\n")
        code = compile(inst_module, '<string>', 'exec')
        return code
    
    def _verify_conditions(self, module, label, expected):
        construct = self.recorder.metadata[module.__name__].constructs[label]
        tag = self.recorder.DEFAULT_TAG
        for i, value in enumerate(expected):
            if expected[i]:
                assert tag in construct.conditions[i], construct.conditions[i]
            else:
                assert not construct.conditions[i], construct.conditions[i]
    
    def test_two_pin_and_t_t(self):
        def test_module():
            a = True
            b = True
            result = a and b
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert True == result
        
        self._verify_conditions(test_module, '3.2', 
                                [True, False, False])
    
    def test_two_pin_and_t_f(self):
        def test_module():
            a = True
            b = False
            result = a and b
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert False == result
        
        self._verify_conditions(test_module, '3.2', 
                                [False, False, True])
    
    def test_two_pin_and_f_t(self):
        def test_module():
            a = False
            b = True
            result = a and b
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert False == result
        
        self._verify_conditions(test_module, '3.2', 
                                [False, True, False])
    
    def test_two_pin_and_f_f(self):
        def test_module():
            a = False
            b = False
            result = a and b
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert False == result
        
        self._verify_conditions(test_module, '3.2', 
                                [False, True, False])
    
    def test_two_pin_or_t_t(self):
        def test_module():
            a = True
            b = True
            result = a or b
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert True == result
        
        self._verify_conditions(test_module, '3.2', 
                                [True, False, False])
    
    def test_two_pin_or_t_f(self):
        def test_module():
            a = True
            b = False
            result = a or b
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert True == result
        
        self._verify_conditions(test_module, '3.2', 
                                [True, False, False])
    
    def test_two_pin_or_f_t(self):
        def test_module():
            a = False
            b = True
            result = a or b
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert True == result
        
        self._verify_conditions(test_module, '3.2', 
                                [False, True, False])
    
    def test_two_pin_or_f_f(self):
        def test_module():
            a = False
            b = False
            result = a or b
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert False == result
        
        self._verify_conditions(test_module, '3.2', 
                                [False, False, True])
    
    def test_three_pin_and_t_t_t(self):
        def test_module():
            a = True
            b = True
            c = True
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert True == result
        
        self._verify_conditions(test_module, '4.2', 
                                [True, False, False, False])
    
    def test_three_pin_and_f_t_t(self):
        def test_module():
            a = False
            b = True
            c = True
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert False == result
        
        self._verify_conditions(test_module, '4.2', 
                                [False, True, False, False])
    
    def test_three_pin_and_t_f_t(self):
        def test_module():
            a = True
            b = False
            c = True
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert False == result
        
        self._verify_conditions(test_module, '4.2', 
                                [False, False, True, False])
    
    def test_three_pin_and_t_t_f(self):
        def test_module():
            a = True
            b = True
            c = False
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert False == result
        
        self._verify_conditions(test_module, '4.2', 
                                [False, False, False, True])
    
    def test_three_pin_and_f_f_f(self):
        def test_module():
            a = False
            b = False
            c = False
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert False == result
        
        self._verify_conditions(test_module, '4.2', 
                                [False, True, False, False])
    
    def test_instrument_if_true_result(self):
        def test_module():
            a = True
            if a:
                result = 1
            else:
                result = 7
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert result is 1
        
        self._verify_conditions(test_module, '2.1', [False, True])
        
    def test_instrument_if_false_result(self):
        def test_module():
            a = False
            if a:
                result = 1
            else:
                result = 7
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert result is 7
        
        self._verify_conditions(test_module, '2.1', [True, False])
        
    def test_instrument_while_false_result(self):
        def test_module():
            a = 0
            result = 0
            while a:
                a -= 1
                result += 1
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert result is 0
        
        self._verify_conditions(test_module, '3.1', [True, False])
        
    def test_instrument_while_with_loop_result(self):
        def test_module():
            a = 3
            result = 0
            while a:
                a -= 1
                result += 1
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert result is 3
        
        self._verify_conditions(test_module, '3.1', [True, True])
            
    def test_instrument_ifexp_true(self):
        def test_module():
            a = 3
            result = 1 if a == 3 else 4
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert result is 1
        
        self._verify_conditions(test_module, '2.1', [False, True])
    
    def test_instrument_ifexp_false(self):
        def test_module():
            a = 3
            result = 1 if a != 3 else 4
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        assert result is 4
        
        self._verify_conditions(test_module, '2.1', [True, False])
    
    def test_instrument_ifexp_true_and_false(self):
        def test_module():
            def test_func(arg):
                return 1 if arg == 3 else 4
            test_func(3)
            test_func(2)
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        
        self._verify_conditions(test_module, '2.1', [True, True])
    
    def test_instrument_ifexp_with_boolop(self):
        def test_module():
            def test_func(a, b):
                return 1 if a and b else 4
            test_func(True, False)
    
        code = self._load_and_compile_module(test_module)
        exec_f(code, globals())
        
        self._verify_conditions(test_module, '2.2', [False, False, True])



import re

from astkit import ast

from instrumental import pragmas

from instrumental.constructs import BooleanDecision
from instrumental.constructs import Comparison
from instrumental.constructs import LogicalAnd
from instrumental.constructs import LogicalOr
from instrumental.constructs import PragmaCondition
from instrumental.constructs import UnreachableCondition


class TestLogicalAnd(object):
    
    def _makeOne(self, selector='3.1', pragmas=[]):
        node = ast.BoolOp(values=[ast.Name(id='a'),
                                  ast.Name(id='b')],
                          op=ast.And(),
                          col_offset=12,
                          lineno=3)
        construct = LogicalAnd('somemodule', selector, node, pragmas)
        return construct
    
    def test_and_as_decision(self):
        construct = self._makeOne()
        assert not construct.is_decision()
    
    def test_and_was_true(self):
        construct = self._makeOne()
        assert not construct.was_true()
    
    def test_and_was_false(self):
        construct = self._makeOne()
        assert not construct.was_false()
    
    def test_decision_result(self):
        construct = self._makeOne()
        expected = """Decision -> somemodule:3.1 < (a and b) >

T ==> 
F ==> """
        assert expected == construct.decision_result(), (expected, construct.decision_result())

    def test_pragma_application(self):
        selector = '3.1'
        pattern = pragmas.no_cond
        pragma_match = re.match(pattern, 'no cond[3.1](T F)')
        pragma = pragmas.PragmaNoCondition(pragma_match)
        construct = self._makeOne(selector='3.1', pragmas=[pragma])
        assert construct.conditions[2] == set([pragma.NO_COND_TAG]),(
            (construct.conditions))

class TestLogicalOr(object):
    
    def _makeOne(self, selector='3.1', pragmas=[]):
        node = ast.BoolOp(values=[ast.Name(id='a'),
                                  ast.Name(id='b')],
                          op=ast.Or(),
                          col_offset=12,
                          lineno=3)
        construct = LogicalOr('somemodule', selector, node, pragmas)
        return construct
    
    def test_or_as_decision(self):
        construct = self._makeOne()
        assert not construct.is_decision()
    
    def test_or_was_true(self):
        construct = self._makeOne()
        assert not construct.was_true()
    
    def test_or_was_false(self):
        construct = self._makeOne()
        assert not construct.was_false()
    
    def test_decision_result(self):
        construct = self._makeOne()
        expected = """Decision -> somemodule:3.1 < (a or b) >

T ==> 
F ==> """
        assert expected == construct.decision_result(), (expected, construct.decision_result())

    def test_pragma_application(self):
        selector = '3.1'
        pattern = pragmas.no_cond
        pragma_match = re.match(pattern, 'no cond[3.1](T *)')
        pragma = pragmas.PragmaNoCondition(pragma_match)
        construct = self._makeOne(selector='3.1', pragmas=[pragma])
        assert construct.conditions[0] == set([pragma.NO_COND_TAG]),(
            (construct.conditions))


class TestBooleanDecision(object):
    
    def _makeOne(self, selector='3.1', pragmas=[]):
        node = ast.Name(id="j",
                        col_offset=12,
                        lineno=3)
        construct = BooleanDecision('somemodule', selector, node, pragmas)
        return construct
        
    def test_decision_result(self):
        construct = self._makeOne()
        expected = """Decision -> somemodule:3.1 < j >

T ==> 
F ==> """
        assert expected == construct.result(), (expected, construct.result())


    def test_description_of_T(self):
        construct = self._makeOne()
        assert 'T' == construct.description(True)

    def test_description_of_F(self):
        construct = self._makeOne()
        assert 'F' == construct.description(False)
    
    def test_was_true_when_it_wasnt(self):
        construct = self._makeOne()
        assert not construct.was_true()
    
    def test_was_true_when_it_was(self):
        construct = self._makeOne()
        construct.conditions[True] = 'X'
        assert construct.was_true()
    
    def test_was_false_when_it_wasnt(self):
        construct = self._makeOne()
        assert not construct.was_false()
    
    def test_was_false_when_it_was(self):
        construct = self._makeOne()
        construct.conditions[False] = 'X'
        assert construct.was_false()
    
    def _makeOneWithLiteral(self):
        node = ast.Name(id="'j'",
                        col_offset=12,
                        lineno=3)
        construct = BooleanDecision('somemodule', '3.1', node, [])
        return construct
        
    def test_number_of_conditions_when_reporting_literals(self):
        construct = self._makeOneWithLiteral()
        assert 2 == construct.number_of_conditions(True)
        
    def test_pragma_application(self):
        selector = '3.1'
        pattern = pragmas.no_cond
        pragma_match = re.match(pattern, 'no cond[3.1](T)')
        pragma = pragmas.PragmaNoCondition(pragma_match)
        construct = self._makeOne(selector='3.1', pragmas=[pragma])
        assert construct.conditions[True] == set([pragma.NO_COND_TAG]),(
            (construct.conditions))
    
    def test_set_unreachable(self):
        construct = self._makeOne()
        construct.set_unreachable(True)
        condition = list(construct.conditions[True])[0]
        assert isinstance(condition, UnreachableCondition)


class TestComparison(object):
    
    def _makeOne(self, selector='3.1', pragmas={}):
        node = ast.Compare(left=ast.Name(id='a'),
                           ops=[ast.Eq()],
                           comparators=[ast.Name(id='b')],
                           col_offset=12,
                           lineno=3)
        construct = Comparison('somemodule', selector, node, pragmas)
        return construct
    
    def test_decision_result(self):
        construct = self._makeOne()
        expected = """Compare -> somemodule:3.1 < (a == b) >

T ==> 
F ==> """
        assert expected == construct.result(), (expected, construct.result())
        
    def test_description_of_T(self):
        construct = self._makeOne()
        assert 'T' == construct.description(True)

    def test_description_of_F(self):
        construct = self._makeOne()
        assert 'F' == construct.description(False)

    def test_was_true_when_it_wasnt(self):
        construct = self._makeOne()
        assert not construct.was_true()
    
    def test_was_true_when_it_was(self):
        construct = self._makeOne()
        construct.conditions[True] = 'X'
        assert construct.was_true()
    
    def test_was_false_when_it_wasnt(self):
        construct = self._makeOne()
        assert not construct.was_false()
    
    def test_was_false_when_it_was(self):
        construct = self._makeOne()
        construct.conditions[False] = 'X'
        assert construct.was_false()

    def _makeOneWithLiteral(self, op=ast.Eq(),
                            left=ast.Num(n=4), right=ast.Num(n=4)):
        node = ast.Compare(left=left,
                           ops=[op],
                           comparators=[right],
                           col_offset=12,
                           lineno=3)
        construct = Comparison('somemodule', '3.1', node, {})
        return construct
    
    def test_number_of_conditions_when_reporting_literals(self):
        construct = self._makeOneWithLiteral()
        assert 2 == construct.number_of_conditions(True)
    
    def test_number_of_conditions_when_not_reporting_literals(self):
        construct = self._makeOneWithLiteral()
        assert 1 == construct.number_of_conditions(False)
    
    def test_pragma_application(self):
        selector = '3.1'
        pattern = pragmas.no_cond
        pragma_match = re.match(pattern, 'no cond[3.1](T)')
        pragma = pragmas.PragmaNoCondition(pragma_match)
        construct = self._makeOne(selector='3.1', pragmas=[pragma])
        assert construct.conditions[True] == set([pragma.NO_COND_TAG]),(
            (construct.conditions))

    def test_set_unreachable(self):
        construct = self._makeOne()
        construct.set_unreachable(True)
        condition = list(construct.conditions[True])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_Eq(self):
        construct = self._makeOneWithLiteral(ast.Eq())
        condition = list(construct.conditions[False])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_NotEq(self):
        construct = self._makeOneWithLiteral(ast.NotEq())
        condition = list(construct.conditions[True])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_Lt(self):
        construct = self._makeOneWithLiteral(ast.Lt())
        condition = list(construct.conditions[True])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_LtE(self):
        construct = self._makeOneWithLiteral(ast.LtE())
        condition = list(construct.conditions[False])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_Gt(self):
        construct = self._makeOneWithLiteral(ast.Gt())
        condition = list(construct.conditions[True])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_GtE(self):
        construct = self._makeOneWithLiteral(ast.GtE())
        condition = list(construct.conditions[False])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_Is(self):
        construct = self._makeOneWithLiteral(ast.Is())
        condition = list(construct.conditions[False])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_IsNot(self):
        construct = self._makeOneWithLiteral(ast.IsNot())
        condition = list(construct.conditions[True])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_In(self):
        construct = self._makeOneWithLiteral(ast.In(), right=ast.List(elts=[]))
        condition = list(construct.conditions[True])[0]
        assert isinstance(condition, UnreachableCondition)
    
    def test_unreachable_NotIn(self):
        construct = self._makeOneWithLiteral(ast.NotIn(), right=ast.List(elts=[]))
        condition = list(construct.conditions[False])[0]
        assert isinstance(condition, UnreachableCondition)
    

class TestUnreachableConditions(object):
    
    def _makeNode(self, op, conditions):
        return ast.BoolOp(values=conditions,
                          op=op)
    
    def _makeAnd(self, *conditions):
        node = self._makeNode(ast.And(), conditions)
        return LogicalAnd('somemodule', '3.1', node, {})
    
    def _makeOr(self, *conditions):
        node = self._makeNode(ast.Or(), conditions)
        return LogicalOr('somemodule', '3.1', node, {})
    
    def _test(self, expected, ctor, conditions):
        construct = ctor(*conditions)
        assert expected == construct.unreachable_conditions(),(
            expected, construct.unreachable_conditions())
    
    def test_And_without_literals(self):
        yield self._test, [], self._makeAnd, (
            ast.Name(id='a'), ast.Name(id='b'), ast.Name(id='c'))
    
    def test_And_with_True_first_pin(self):
        yield self._test, [1], self._makeAnd, (
            ast.Name(id='True'), ast.Name(id='b'), ast.Name(id='c'))
    
    def test_And_with_False_first_pin(self):
        yield self._test, [], self._makeAnd, (
            ast.Name(id='False'), ast.Name(id='b'), ast.Name(id='c'))
    
    def test_And_with_True_second_pin(self):
        yield self._test, [2], self._makeAnd, (
            ast.Name(id='a'), ast.Name(id='True'), ast.Name(id='c'))
    
    def test_And_with_False_second_pin(self):
        yield self._test, [], self._makeAnd, (
            ast.Name(id='a'), ast.Name(id='False'), ast.Name(id='c'))
    
    def test_And_with_True_third_pin(self):
        yield self._test, [3], self._makeAnd, (
            ast.Name(id='a'), ast.Name(id='b'), ast.Name(id='True'))
    
    def test_And_with_False_third_pin(self):
        yield self._test, [0], self._makeAnd, (
            ast.Name(id='a'), ast.Name(id='b'), ast.Name(id='False'))
    
    def test_Or_without_literals(self):
        yield self._test, [], self._makeOr, (
            ast.Name(id='a'), ast.Name(id='b'), ast.Name(id='c'))
    
    def test_Or_with_True_first_pin(self):
        yield self._test, [], self._makeOr, (
            ast.Name(id='True'), ast.Name(id='b'), ast.Name(id='c'))
    
    def test_Or_with_False_first_pin(self):
        yield self._test, [0], self._makeOr, (
            ast.Name(id='False'), ast.Name(id='b'), ast.Name(id='c'))
    
    def test_Or_with_True_second_pin(self):
        yield self._test, [], self._makeOr, (
            ast.Name(id='a'), ast.Name(id='True'), ast.Name(id='c'))
    
    def test_Or_with_False_second_pin(self):
        yield self._test, [1], self._makeOr, (
            ast.Name(id='a'), ast.Name(id='False'), ast.Name(id='c'))
    
    def test_Or_with_True_third_pin(self):
        yield self._test, [3], self._makeOr, (
            ast.Name(id='a'), ast.Name(id='b'), ast.Name(id='True'))
    
    def test_Or_with_False_third_pin(self):
        yield self._test, [2], self._makeOr, (
            ast.Name(id='a'), ast.Name(id='b'), ast.Name(id='False'))
    
class TestPragmaCondition(object):
    
    def test_str(self):
        assert 'P' == str(PragmaCondition())



from astkit import ast

from instrumental.compat import exec_f
from instrumental.recorder import ExecutionRecorder
from instrumental.test import DummyConfig
from instrumental.test import InstrumentationTestCase

class TestPragmaFinder(object):
    
    def setup(self):
        from instrumental.pragmas import PragmaFinder
        self.finder = PragmaFinder()
    
    def test_pragma_no_cover(self):
        from instrumental.pragmas import PragmaNoCover
        source = """
acc = 1
acc += 2
if add_three:
    acc += 3 # pragma: no cover
acc += 4
"""
        pragmas = self.finder.find_pragmas(source)
        assert 6 == len(pragmas), pragmas
        assert not pragmas[1]
        assert not pragmas[2]
        assert not pragmas[3]
        assert not pragmas[4]
        assert pragmas[5], pragmas
        assert isinstance(list(pragmas[5])[0], PragmaNoCover)
        assert not pragmas[6]
    
    def test_pragma_nocover(self):
        from instrumental.pragmas import PragmaNoCover
        source = """
acc = 1
acc += 2
if add_three:
    acc += 3 # pragma: nocover
acc += 4
"""
        pragmas = self.finder.find_pragmas(source)
        assert 6 == len(pragmas), pragmas
        assert not pragmas[1]
        assert not pragmas[2]
        assert not pragmas[3]
        assert not pragmas[4]
        assert pragmas[5], pragmas
        assert isinstance(list(pragmas[5])[0], PragmaNoCover)
        assert not pragmas[6]
    
    def test_pragma_no_cover_on_FunctionDefn(self):
        from instrumental.pragmas import PragmaNoCover
        source = """
def somefunc(args): # pragma: nocover
    return 'howdy'
"""
        pragmas = self.finder.find_pragmas(source)
        assert 3 == len(pragmas), pragmas
        assert not pragmas[1]
        assert pragmas[2], pragmas
        assert isinstance(list(pragmas[2])[0], PragmaNoCover)
        assert pragmas[3], pragmas
        assert isinstance(list(pragmas[3])[0], PragmaNoCover)
    
    def test_pragma_no_cond_T(self):
        from instrumental.pragmas import PragmaNoCondition
        source = """
acc = 1
acc += 2
if add_three: # pragma: no cond(T)
    acc += 3
acc += 4
"""
        pragmas = self.finder.find_pragmas(source)
        assert 6 == len(pragmas), pragmas
        assert not pragmas[1]
        assert not pragmas[2]
        assert not pragmas[3]
        assert 1 == len(pragmas[4])
        pragma_4 = list(pragmas[4])[0]
        assert isinstance(pragma_4, PragmaNoCondition)
        assert pragma_4.conditions == ['T']
        assert not pragmas[5]
        assert not pragmas[6]

    def test_pragma_no_cond_T_F(self):
        from instrumental.pragmas import PragmaNoCondition
        source = """
acc = 1
acc += 2
if add_three and add_four: # pragma: no cond(T F)
    acc += 3
acc += 4
"""
        pragmas = self.finder.find_pragmas(source)
        assert 6 == len(pragmas), pragmas
        assert not pragmas[1]
        assert not pragmas[2]
        assert not pragmas[3]
        assert 1 == len(pragmas[4])
        pragma_4 = list(pragmas[4])[0]
        assert isinstance(pragma_4, PragmaNoCondition)
        assert pragma_4.conditions == ['T F']
        assert not pragmas[5]
        assert not pragmas[6]

    def test_pragma_no_cond_multiple_conditions(self):
        from instrumental.pragmas import PragmaNoCondition
        source = """
acc = 1
acc += 2
if add_three and add_four: # pragma: no cond(T F,F T)
    acc += 3
acc += 4
"""
        pragmas = self.finder.find_pragmas(source)
        assert 6 == len(pragmas), pragmas
        assert not pragmas[1]
        assert not pragmas[2]
        assert not pragmas[3]
        assert 1 == len(pragmas[4])
        pragma_4 = list(pragmas[4])[0]
        assert isinstance(pragma_4, PragmaNoCondition)
        assert sorted(pragma_4.conditions) == ['F T', 'T F']
        assert not pragmas[5]
        assert not pragmas[6]
    
    def test_pragma_no_cond_with_condition_selector(self):
        from instrumental.pragmas import PragmaNoCondition
        source = """
acc = 1
acc += 2
tot, err = a and b, c or d # pragma: no cond[.2](T F)
assert tot or err
"""
        pragmas = self.finder.find_pragmas(source)
        assert 5 == len(pragmas), pragmas
        assert not pragmas[1]
        assert not pragmas[2]
        assert not pragmas[3]
        assert 1 == len(pragmas[4])
        pragma_4_2 = list(pragmas[4])[0]
        assert isinstance(pragma_4_2, PragmaNoCondition)
        assert pragma_4_2.conditions == ['T F']
        assert not pragmas[5]
    
    def test_pragma_no_cond_with_full_selector(self):
        from instrumental.pragmas import PragmaNoCondition
        source = """# pragma: no cond[4.2](T F)
acc = 1
acc += 2
tot, err = a and b, c or d
assert tot or err
"""
        pragmas = self.finder.find_pragmas(source)
        assert 5 == len(pragmas), pragmas
        assert not pragmas[1]
        assert not pragmas[2]
        assert not pragmas[3]
        assert 1 == len(pragmas[4])
        pragma_4_2 = list(pragmas[4])[0]
        assert isinstance(pragma_4_2, PragmaNoCondition)
        assert pragma_4_2.selector == '2'
        assert pragma_4_2.conditions == ['T F']
        assert not pragmas[5]
    
class TestPragmaNoCondition(InstrumentationTestCase):
    
    def setup(self):
        self.config = DummyConfig()
        ExecutionRecorder.reset()
        ExecutionRecorder.get().start()
    
    def teardown(self):
        ExecutionRecorder.get().stop()        
    
    @property
    def recorder(self):
        return ExecutionRecorder.get()
    
    def test_conditions_are_ignored(self):
        import re
        from astkit import ast
        from instrumental.constructs import LogicalAnd
        from instrumental.pragmas import no_cond
        from instrumental.pragmas import PragmaNoCondition
        node = ast.BoolOp(values=[ast.Name(id="x"), ast.Name(id="y")],
                          op=ast.And(),
                          lineno=17,
                          col_offset=1)
        match = re.match(no_cond, 'no cond(T F,F *)')
        pragma = PragmaNoCondition(match)
        construct = LogicalAnd('<string>', '17.1', node, set([pragma]))
        assert '(x and y)' == construct.source
        assert 3 == construct.number_of_conditions(False)
        assert "T T" == construct.description(0)
        assert "F *" == construct.description(1)
        assert "T F" == construct.description(2)
        
        # T T
        construct.record(True, 0, '*')
        construct.record(True, 1, '*')
        
        assert not construct.conditions_missed(False)
        assert construct.conditions[0] == set(['*'])
        assert construct.conditions[1] == set(['P'])
        assert construct.conditions[2] == set(['P'])
    
    def test_roundtrip_with_full_selector(self):
        def dummy_module():
            # pragma: no cond[5.3](F *)
            a = True
            b = False
            c = False
            e = a and b or c
        mod = self._instrument_module(dummy_module)
        code = compile(mod, '<string>', 'exec')
        exec_f(code, globals())
        assert not e
        
        construct = self.recorder.metadata['dummy_module'].constructs['5.3']
        assert construct.source == '(a and b)', construct.source
        assert construct.conditions_missed(False)
        assert construct.conditions[0] == set([]) # T T
        assert construct.conditions[1] == set(['P']) # F *
        assert construct.conditions[2] == set([self.recorder.DEFAULT_TAG]) # T F
        
    def test_roundtrip_with_expression_selector(self):
        def dummy_module():
            a = True
            b = False
            c = False
            e = a and b or c # pragma: no cond[.3](F *)
        mod = self._instrument_module(dummy_module)
        code = compile(mod, '<string>', 'exec')
        exec_f(code, globals())
        assert not e
        
        construct = self.recorder.metadata['dummy_module'].constructs['4.3']
        assert construct.source == '(a and b)'
        assert construct.conditions_missed(False)
        assert construct.conditions[0] == set([]) # T T
        assert construct.conditions[1] == set(['P']) # F *
        assert construct.conditions[2] == set([self.recorder.DEFAULT_TAG]) # T F
        
    def test_conditions_are_ignored_for_decision(self):
        import re
        from astkit import ast
        from instrumental.constructs import BooleanDecision
        from instrumental.pragmas import no_cond
        from instrumental.pragmas import PragmaNoCondition
        node = ast.Name(id="x",
                        lineno=17,
                        col_offset=1)
        match = re.match(no_cond, 'no cond(T)')
        pragma = PragmaNoCondition(match)
        construct = BooleanDecision('<string>', '17.1', node, set([pragma]))
        assert 'x' == construct.source
        assert 1 == construct.number_of_conditions(False)
        assert "T" == construct.description(True)
        assert "F" == construct.description(False)
        
        # T T
        construct.record(False, '*')
        
        assert not construct.conditions_missed(False)
        assert construct.conditions[True] == set(['P'])
        assert construct.conditions[False] == set(['*'])
    
class TestInstrumentationWithPragmas(InstrumentationTestCase):
    
    def test_ClassDef(self):
        def test_module():
            foo = 7
            class FooClass(object): # pragma: no cover
                bar = 4
            baz = 8
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        
        assert isinstance(inst_module.body[3], ast.Assign)
        assert isinstance(inst_module.body[3].targets[0], ast.Name)
        assert inst_module.body[3].targets[0].id == 'foo'
        assert isinstance(inst_module.body[3].value, ast.Num)
        assert inst_module.body[3].value.n == 7
        
        assert isinstance(inst_module.body[4], ast.ClassDef), inst_module.body[4]
        assert inst_module.body[4].name == 'FooClass'
        assert isinstance(inst_module.body[4].bases[0], ast.Name)
        assert inst_module.body[4].bases[0].id == 'object'
        
        assert 1 == len(inst_module.body[4].body)
        assert isinstance(inst_module.body[4].body[0], ast.Assign)
        assert isinstance(inst_module.body[4].body[0].targets[0], ast.Name)
        assert inst_module.body[4].body[0].targets[0].id == 'bar'
        assert isinstance(inst_module.body[4].body[0].value, ast.Num)
        assert inst_module.body[4].body[0].value.n == 4
        
        self._assert_record_statement(inst_module.body[5], 'test_module', 4)
        
        assert isinstance(inst_module.body[6], ast.Assign)
        assert isinstance(inst_module.body[6].targets[0], ast.Name)
        assert inst_module.body[6].targets[0].id == 'baz'
        assert isinstance(inst_module.body[6].value, ast.Num)
        assert inst_module.body[6].value.n == 8

    def test_FunctionDef(self):
        def test_module():
            foo = 7
            def foo_func(): # pragma: no cover
                bar = 4
            baz = 8
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        
        assert isinstance(inst_module.body[3], ast.Assign)
        assert isinstance(inst_module.body[3].targets[0], ast.Name)
        assert inst_module.body[3].targets[0].id == 'foo'
        assert isinstance(inst_module.body[3].value, ast.Num)
        assert inst_module.body[3].value.n == 7
        
        assert isinstance(inst_module.body[4], ast.FunctionDef), inst_module.body[4]
        assert inst_module.body[4].name == 'foo_func'
        assert not inst_module.body[4].args.args
        assert not inst_module.body[4].args.kwarg
        assert not inst_module.body[4].decorator_list
        
        assert 1 == len(inst_module.body[4].body)
        assert isinstance(inst_module.body[4].body[0], ast.Assign)
        assert isinstance(inst_module.body[4].body[0].targets[0], ast.Name)
        assert inst_module.body[4].body[0].targets[0].id == 'bar'
        assert isinstance(inst_module.body[4].body[0].value, ast.Num)
        assert inst_module.body[4].body[0].value.n == 4
        
        self._assert_record_statement(inst_module.body[5], 'test_module', 4)
        
        assert isinstance(inst_module.body[6], ast.Assign)
        assert isinstance(inst_module.body[6].targets[0], ast.Name)
        assert inst_module.body[6].targets[0].id == 'baz'
        assert isinstance(inst_module.body[6].value, ast.Num)
        assert inst_module.body[6].value.n == 8



from instrumental.instrument import force_location

class FakeNode(object):
    _fields = []

class TestForceLocation(object):
    
    def test_force_single_node(object):
        node = FakeNode()
        node.lineno = 5
        node.col_offset = 28
        
        force_location(node, 3)
        
        assert 3 == node.lineno
        assert 0 == node.col_offset



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import imp
import logging
import os
import re
import sys

from astkit import ast
from astkit.render import SourceCodeRenderer

from instrumental.compat import exec_f

log = logging.getLogger(__name__)

class ModuleLoader(object):
    
    def __init__(self, fullpath, visitor_factory):
        self.fullpath = fullpath
        self.visitor_factory = visitor_factory
    
    def _get_source(self, path):
        """ Get the source code from a file path """
        with open(path, 'r') as f:
            source = f.read()
        return source
    
    def _get_code(self, fullname):
        """ Get the instrumented code for a module
            
            Given a dotted module name, return a tuple that looks like
            (<is the module a package>, <module's code object instrumented>)
        """
        # packages are loaded from __init__.py files
        ispkg = self.fullpath.endswith('__init__.py')
        code_str = self._get_source(self.fullpath)
        visitor = self.visitor_factory.create(fullname, code_str)
        code_tree = ast.parse(code_str)
        new_code_tree = visitor.visit(code_tree)
        log.debug(SourceCodeRenderer.render(new_code_tree))
        code = compile(new_code_tree, self.fullpath, 'exec')
        return (ispkg, code)
    
    def load_module(self, fullname):
        ispkg, code = self._get_code(fullname)
        mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
        mod.__file__ = self.fullpath
        mod.__loader__ = self
        if ispkg:
            mod.__path__ = [os.path.dirname(self.fullpath)]
        exec_f(code, mod.__dict__)
        return mod

class ImportHook(object):
    """ An implementation of an import hook per PEP 302
    """
    
    def __init__(self, target, ignores, visitor_factory):
        self.target = target
        self.ignores = ignores
        self.visitor_factory = visitor_factory
    
    def find_module(self, fullname, path=[]):
        log.debug("find_module(%s, path=%r)",fullname, path)
        if ((not re.match(self.target, fullname)) 
            or
            any([re.match(ignore, fullname) for ignore in self.ignores])):
            return None
        
        if not path:
            path = sys.path
        
        for directory in path:
            loader = self._loader_for_path(directory, fullname)
            if loader:
                return loader
    
    def _loader_for_path(self, directory, fullname):
        log.debug("[loader for path] directory: %s; fullname: %s",
                  directory, fullname)
        module_path = os.path.join(directory, fullname.split('.')[-1]) + ".py"
        if os.path.exists(module_path):
            log.debug("loading module from %s", module_path)
            loader = ModuleLoader(module_path, self.visitor_factory)
            return loader
        
        package_path = os.path.join(directory, fullname.split('.')[-1], '__init__.py')
        if os.path.exists(package_path):
            loader = ModuleLoader(package_path, self.visitor_factory)
            return loader
        



import inspect
import sys

from astkit import ast
from astkit.render import SourceCodeRenderer as renderer

from instrumental.test import DummyConfig
from instrumental.test import InstrumentationTestCase
from instrumental.test import load_module

class TestInstrumentNodesPython2(InstrumentationTestCase):
    
    def test_simple_module(self):
        def test_module():
            a = True
            b = True
            result = a and b
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
    def test_FunctionDef(self):
        def test_module():
            def foo():
                bar = 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.FunctionDef)
        assert inst_module.body[3].name == 'foo'
        assert isinstance(inst_module.body[3].args, ast.arguments)
        assert not inst_module.body[3].args.args
        assert not inst_module.body[3].args.vararg
        assert not inst_module.body[3].args.kwarg
        assert not inst_module.body[3].args.defaults
        assert 2 == len(inst_module.body[3].body)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Assign)
        assert isinstance(inst_module.body[3].body[1].targets[0], ast.Name)
        assert inst_module.body[3].body[1].targets[0].id == 'bar'
        assert isinstance(inst_module.body[3].body[1].value, ast.Num)
        assert inst_module.body[3].body[1].value.n == 4
    
    def test_FunctionDef_with_docstring(self):
        def test_module():
            def foo():
                "I'm a little docstring"
                bar = 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.FunctionDef)
        assert inst_module.body[3].name == 'foo'
        assert isinstance(inst_module.body[3].args, ast.arguments)
        assert not inst_module.body[3].args.args
        assert not inst_module.body[3].args.vararg
        assert not inst_module.body[3].args.kwarg
        assert not inst_module.body[3].args.defaults
        assert 3 == len(inst_module.body[3].body)
        assert isinstance(inst_module.body[3].body[0], ast.Expr)
        assert isinstance(inst_module.body[3].body[0].value, ast.Str)
        assert inst_module.body[3].body[0].value.s == "I'm a little docstring"
        self._assert_record_statement(inst_module.body[3].body[1], 'test_module', 3)
        assert isinstance(inst_module.body[3].body[2], ast.Assign)
        assert isinstance(inst_module.body[3].body[2].targets[0], ast.Name)
        assert inst_module.body[3].body[2].targets[0].id == 'bar'
        assert isinstance(inst_module.body[3].body[2].value, ast.Num)
        assert inst_module.body[3].body[2].value.n == 4
    
    def test_FunctionDef_with_docstring_and_no_cover(self):
        def test_module():
            def foo(): # pragma: no cover
                "I'm a little docstring"
                bar = 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        assert isinstance(inst_module.body[2], ast.FunctionDef)
        assert inst_module.body[2].name == 'foo'
        assert isinstance(inst_module.body[2].args, ast.arguments)
        assert not inst_module.body[2].args.args
        assert not inst_module.body[2].args.vararg
        assert not inst_module.body[2].args.kwarg
        assert not inst_module.body[2].args.defaults
        assert 2 == len(inst_module.body[2].body)
        assert isinstance(inst_module.body[2].body[0], ast.Expr)
        assert isinstance(inst_module.body[2].body[0].value, ast.Str)
        assert inst_module.body[2].body[0].value.s == "I'm a little docstring"
        assert isinstance(inst_module.body[2].body[1], ast.Assign)
        assert isinstance(inst_module.body[2].body[1].targets[0], ast.Name)
        assert inst_module.body[2].body[1].targets[0].id == 'bar'
        assert isinstance(inst_module.body[2].body[1].value, ast.Num)
        assert inst_module.body[2].body[1].value.n == 4
    
    def test_ClassDef(self):
        def test_module():
            class FooClass(object):
                bar = 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.ClassDef)
        assert inst_module.body[3].name == 'FooClass'
        assert isinstance(inst_module.body[3].bases[0], ast.Name)
        assert inst_module.body[3].bases[0].id == 'object'
        assert 2 == len(inst_module.body[3].body)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Assign)
        assert isinstance(inst_module.body[3].body[1].targets[0], ast.Name)
        assert inst_module.body[3].body[1].targets[0].id == 'bar'
        assert isinstance(inst_module.body[3].body[1].value, ast.Num)
        assert inst_module.body[3].body[1].value.n == 4
    
    def test_ClassDef_with_docstring(self):
        def test_module():
            class FooClass(object):
                "I'm a little docstring"
                bar = 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.ClassDef)
        assert inst_module.body[3].name == 'FooClass'
        assert isinstance(inst_module.body[3].bases[0], ast.Name)
        assert inst_module.body[3].bases[0].id == 'object'
        assert 3 == len(inst_module.body[3].body)
        assert isinstance(inst_module.body[3].body[0], ast.Expr)
        assert isinstance(inst_module.body[3].body[0].value, ast.Str)
        assert inst_module.body[3].body[0].value.s == "I'm a little docstring"
        self._assert_record_statement(inst_module.body[3].body[1], 'test_module', 3)
        assert isinstance(inst_module.body[3].body[2], ast.Assign)
        assert isinstance(inst_module.body[3].body[2].targets[0], ast.Name)
        assert inst_module.body[3].body[2].targets[0].id == 'bar'
        assert isinstance(inst_module.body[3].body[2].value, ast.Num)
        assert inst_module.body[3].body[2].value.n == 4
    
    def test_Return(self):
        def test_module():
            def foo():
                return 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.FunctionDef)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Return)
        assert isinstance(inst_module.body[3].body[1].value, ast.Num)
        assert inst_module.body[3].body[1].value.n == 4
    
    def test_Delete(self):
        def test_module():
            def foo():
                del bar
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.FunctionDef)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Delete)
        assert isinstance(inst_module.body[3].body[1].targets[0], ast.Name)
        assert inst_module.body[3].body[1].targets[0].id == 'bar'
    
    def test_Assert(self):
        def test_module():
            assert True
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.Assert)
        assert isinstance(inst_module.body[3].test, ast.Name)
        assert inst_module.body[3].test.id == 'True'
    
    def test_Assign(self):
        def test_module():
            a = True
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.Assign)
        assert isinstance(inst_module.body[3].targets[0], ast.Name)
        assert inst_module.body[3].targets[0].id == 'a'
        assert isinstance(inst_module.body[3].value, ast.Name)
        assert inst_module.body[3].value.id == 'True'
    
    def test_AugAssign(self):
        def test_module():
            a += 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.AugAssign)
        assert isinstance(inst_module.body[3].target, ast.Name)
        assert inst_module.body[3].target.id == 'a'
        assert isinstance(inst_module.body[3].op, ast.Add)
        assert isinstance(inst_module.body[3].value, ast.Num)
        assert inst_module.body[3].value.n == 4
    
    def test_Break(self):
        def test_module():
            while True:
                break
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.While)
        assert isinstance(inst_module.body[3].test, ast.Call)
        assert isinstance(inst_module.body[3].test.func, ast.Attribute)
        assert isinstance(inst_module.body[3].test.func.value, ast.Name)
        assert inst_module.body[3].test.func.value.id == '_xxx_recorder_xxx_'
        assert inst_module.body[3].test.func.attr == 'record'
        assert isinstance(inst_module.body[3].test.args[0], ast.Name)
        assert inst_module.body[3].test.args[0].id == 'True'
        assert isinstance(inst_module.body[3].test.args[1], ast.Str)
        assert inst_module.body[3].test.args[1].s == 'test_module'
        assert isinstance(inst_module.body[3].test.args[2], ast.Str)
        assert inst_module.body[3].test.args[2].s == '1.1'
        assert not inst_module.body[3].test.keywords
        assert not hasattr(inst_module.body[3].test, 'starargs')
        assert not hasattr(inst_module.body[3].test, 'kwargs')
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Break)
    
    def test_Continue(self):
        def test_module():
            while True:
                continue
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.While)
        assert isinstance(inst_module.body[3].test, ast.Call)
        assert isinstance(inst_module.body[3].test.func, ast.Attribute)
        assert isinstance(inst_module.body[3].test.func.value, ast.Name)
        assert inst_module.body[3].test.func.value.id == '_xxx_recorder_xxx_'
        assert inst_module.body[3].test.func.attr == 'record'
        assert isinstance(inst_module.body[3].test.args[0], ast.Name)
        assert inst_module.body[3].test.args[0].id == 'True'
        assert isinstance(inst_module.body[3].test.args[1], ast.Str)
        assert inst_module.body[3].test.args[1].s == 'test_module'
        assert isinstance(inst_module.body[3].test.args[2], ast.Str)
        assert inst_module.body[3].test.args[2].s == '1.1'
        assert not inst_module.body[3].test.keywords
        assert not hasattr(inst_module.body[3].test, 'starargs')
        assert not hasattr(inst_module.body[3].test, 'kwargs')
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Continue)
    
    if sys.version_info[0] < 3:
        from instrumental.test.py2_only import test_Print
    
    def test_For(self):
        def test_module():
            for i in [1,2,3,5]:
                return i
            else:
                return 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.For)
        assert isinstance(inst_module.body[3].target, ast.Name)
        assert inst_module.body[3].target.id == 'i'
        assert isinstance(inst_module.body[3].iter, ast.List)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1],ast.Return)
        self._assert_record_statement(inst_module.body[3].orelse[0], 'test_module', 4)
        assert isinstance(inst_module.body[3].orelse[1],ast.Return)
    
    def test_For_with_no_cover(self):
        def test_module():
            for i in [1,2,3,5]: # pragma: no cover
                return i
            else:
                return 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        assert isinstance(inst_module.body[2], ast.For)
        assert isinstance(inst_module.body[2].target, ast.Name)
        assert inst_module.body[2].target.id == 'i'
        assert isinstance(inst_module.body[2].iter, ast.List)
        assert isinstance(inst_module.body[2].body[0],ast.Return)
        assert isinstance(inst_module.body[2].orelse[0],ast.Return)
    
    def test_Global(self):
        def test_module():
            global frog
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.Global)
    
    def test_While(self):
        def test_module():
            while i:
                return i
            else:
                return 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.While)
        assert isinstance(inst_module.body[3].test, ast.Call)
        assert isinstance(inst_module.body[3].test.func, ast.Attribute)
        assert isinstance(inst_module.body[3].test.func.value, ast.Name)
        assert inst_module.body[3].test.func.value.id == '_xxx_recorder_xxx_'
        assert inst_module.body[3].test.func.attr == 'record'
        assert isinstance(inst_module.body[3].test.args[0], ast.Name)
        assert inst_module.body[3].test.args[0].id == 'i'
        assert isinstance(inst_module.body[3].test.args[1], ast.Str)
        assert inst_module.body[3].test.args[1].s == 'test_module'
        assert isinstance(inst_module.body[3].test.args[2], ast.Str)
        assert inst_module.body[3].test.args[2].s == '1.1'
        assert not inst_module.body[3].test.keywords
        assert not hasattr(inst_module.body[3].test, 'starargs')
        assert not hasattr(inst_module.body[3].test, 'kwargs')
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1],ast.Return)
        self._assert_record_statement(inst_module.body[3].orelse[0], 'test_module', 4)
        assert isinstance(inst_module.body[3].orelse[1],ast.Return)

    def test_While_with_no_cover(self):
        def test_module():
            while i: # pragma: no cover
                return i
            else:
                return 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        assert isinstance(inst_module.body[2], ast.While)
        assert isinstance(inst_module.body[2].body[0], ast.Return)
        assert isinstance(inst_module.body[2].orelse[0], ast.Return)

    def test_While(self):
        def test_module():
            while i:
                return i
            else:
                return 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.While)
        assert isinstance(inst_module.body[3].test, ast.Call)
        assert isinstance(inst_module.body[3].test.func, ast.Attribute)
        assert isinstance(inst_module.body[3].test.func.value, ast.Name)
        assert inst_module.body[3].test.func.value.id == '_xxx_recorder_xxx_'
        assert inst_module.body[3].test.func.attr == 'record'
        assert isinstance(inst_module.body[3].test.args[0], ast.Name)
        assert inst_module.body[3].test.args[0].id == 'i'
        assert isinstance(inst_module.body[3].test.args[1], ast.Str)
        assert inst_module.body[3].test.args[1].s == 'test_module'
        assert isinstance(inst_module.body[3].test.args[2], ast.Str)
        assert inst_module.body[3].test.args[2].s == '1.1'
        assert not inst_module.body[3].test.keywords
        assert not hasattr(inst_module.body[3].test, 'starargs')
        assert not hasattr(inst_module.body[3].test, 'kwargs')
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1],ast.Return)
        self._assert_record_statement(inst_module.body[3].orelse[0], 'test_module', 4)
        assert isinstance(inst_module.body[3].orelse[1],ast.Return)

    def test_With(self):
        def test_module():
            with i as x:
                foo(x)
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Expr)
        assert isinstance(inst_module.body[3].body[1].value, ast.Call)
    
    def test_If_with_no_cover(self):
        def test_module():
            if i: # pragma: no cover
                return i
            else:
                return 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        assert isinstance(inst_module.body[2], ast.If)
        assert isinstance(inst_module.body[2].body[0],ast.Return)
        assert isinstance(inst_module.body[2].orelse[0],ast.Return)
    
    def test_If_with_no_cover_else(self):
        def test_module():
            if i:
                return i
            else: # pragma: no cover
                return 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.If)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1],ast.Return)
        assert isinstance(inst_module.body[3].orelse[0],ast.Return)
    
    def test_ImportFrom(self):
        def test_module():
            from flag import pole
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.ImportFrom)
    
    def test_ImportFrom_with_no_cover(self):
        def test_module():
            from flag import pole # pragma: no cover
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        assert isinstance(inst_module.body[2], ast.ImportFrom)
    
    def test_Module_with_docstring(self):
        def test_module():
            """ I am a docstring
                I span more than one line """
            if u:
                return u
            else:
                return 'else'
        inst_module = self._instrument_module(test_module)
        assert ast.get_docstring(inst_module)
        self._assert_recorder_setup(inst_module, 1)
    
    def test_Module_with_future_import(self):
        from instrumental.instrument import CoverageAnnotator
        from instrumental.metadata import MetadataGatheringVisitor
        from instrumental.pragmas import PragmaFinder
        
        source = """from __future__ import with_statement
if u:
    return u
else:
    return 'else'
"""
        module = ast.parse(source)
        pragmas = PragmaFinder().find_pragmas(source)
        config = DummyConfig()
        metadata = MetadataGatheringVisitor.analyze(config,
                                                    'somemodule', 
                                                    'somemodule.py',
                                                    source, pragmas)
        self.recorder.add_metadata(metadata)
        transformer = CoverageAnnotator(config,
                                        'somemodule',
                                        self.recorder)
        inst_module = transformer.visit(module)
        
        assert isinstance(inst_module.body[0], ast.ImportFrom)
        assert module.body[0].module == '__future__'
        assert isinstance(module.body[0].names[0], ast.alias)
        assert module.body[0].names[0].name == 'with_statement'
        self._assert_recorder_setup(inst_module, 1)
    
    def test_Module_with_docstring_and_future_import(self):
        from instrumental.instrument import CoverageAnnotator
        from instrumental.metadata import MetadataGatheringVisitor
        from instrumental.pragmas import PragmaFinder
        
        source = """'a docstring am I'
from __future__ import with_statement
if u:
    return u
else:
    return 'else'
"""
        module = ast.parse(source)
        pragmas = PragmaFinder().find_pragmas(source)
        config = DummyConfig()
        metadata = MetadataGatheringVisitor.analyze(config,
                                                    'somemodule', 
                                                    'somemodule.py',
                                                    source, pragmas)
        self.recorder.add_metadata(metadata)
        transformer = CoverageAnnotator(config,
                                        'somemodule',
                                        self.recorder)
        inst_module = transformer.visit(module)
        
        
        assert isinstance(inst_module.body[1], ast.ImportFrom)
        assert module.body[1].module == '__future__'
        assert isinstance(module.body[1].names[0], ast.alias)
        assert module.body[1].names[0].name == 'with_statement'
        self._assert_recorder_setup(inst_module, 2)
    
    def test_ExceptHandler(self):
        def test_module():
            try:
                return u
            except Exception, exc:
                return 'other'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module, 0)
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.TryExcept)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Return)
        self._assert_record_statement(inst_module.body[3].handlers[0].body[0], 'test_module', 4)
        assert isinstance(inst_module.body[3].handlers[0].body[1], ast.Return)
        
    def test_ExceptHandler_with_no_cover(self):
        def test_module():
            try:
                return u
            except Exception, exc: # pragma: no cover
                return 'other'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module, 0)
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.TryExcept)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Return)
        assert isinstance(inst_module.body[3].handlers[0].body[0], ast.Return)
        
    def test_Pass(self):
        def test_module():
            pass
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.Pass)
    
    def test_Raise(self):
        def test_module():
            raise Exception("no problem")
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.Raise)
    
    def test_TryFinally(self):
        def test_module():
            try:
                return u
            finally:
                foo()
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module, 0)
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.TryFinally)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Return)
        self._assert_record_statement(inst_module.body[3].finalbody[0], 'test_module', 4)
        assert isinstance(inst_module.body[3].finalbody[1], ast.Expr)
        assert isinstance(inst_module.body[3].finalbody[1].value, ast.Call)
        



import imp
import os
import sys

from instrumental.test import DummyConfig

class TestMonkeyPatch(object):
    
    def test_monkeypatch(self):
        import imp
        from instrumental.instrument import AnnotatorFactory
        from instrumental.monkey import monkeypatch_imp
        
        original_load_module = imp.load_module
        config = DummyConfig()
        monkeypatch_imp([], [], AnnotatorFactory(config, None))
        assert original_load_module != imp.load_module

class TestLoadModule(object):
    
    def setup(self):
        from instrumental.recorder import ExecutionRecorder
        ExecutionRecorder.reset()
        self._pre_test_modules = sys.modules.keys()
        self.config = DummyConfig()
    
    def teardown(self):
        _post_test_modules = sys.modules.keys()
        for modname in _post_test_modules:
            if modname not in self._pre_test_modules:
                if modname.startswith('instrumental.test.samples'):
                    del sys.modules[modname]
    
    def test_load_non_target_module(self):
        from instrumental.instrument import AnnotatorFactory
        from instrumental.monkey import load_module_factory
        from instrumental.recorder import ExecutionRecorder
        
        recorder = ExecutionRecorder.get()
        visitor_factory = AnnotatorFactory(self.config, recorder)
        load_module = load_module_factory([], 
                                          [],
                                          visitor_factory)
        
        import instrumental.test.samples
        samples_directory = os.path.dirname(instrumental.test.samples.__file__)
        simple_name = 'instrumental.test.samples.simple'
        simple_path = os.path.join(samples_directory, 'simple.py')
        simple_fh = open(simple_path, 'r')
        load_module(simple_name,
                    simple_fh,
                    simple_path,
                    ('.py', 'r', imp.PY_SOURCE)
                    )
        
        assert simple_name in sys.modules
    
    def test_load_module(self):
        from instrumental.instrument import AnnotatorFactory
        from instrumental.metadata import MetadataGatheringVisitor
        from instrumental.monkey import load_module_factory
        from instrumental.pragmas import PragmaFinder
        from instrumental.recorder import ExecutionRecorder
        
        import instrumental.test.samples
        samples_directory = os.path.dirname(instrumental.test.samples.__file__)
        simple_name = 'instrumental.test.samples.simple'
        simple_path = os.path.join(samples_directory, 'simple.py')
        simple_fh = open(simple_path, 'r')
        
        source = open(simple_path, "r").read()
        pragmas = PragmaFinder().find_pragmas(source)
        metadata = MetadataGatheringVisitor.analyze(self.config,
                                                    simple_name, 
                                                    'simple_name.py', 
                                                    source, 
                                                    pragmas)
        
        recorder = ExecutionRecorder.get()
        recorder.add_metadata(metadata)
        visitor_factory = AnnotatorFactory(self.config, recorder)
        load_module = load_module_factory(['instrumental.test.samples.simple'], 
                                          [],
                                          visitor_factory)
        
        load_module(simple_name,
                    simple_fh,
                    simple_path,
                    ('.py', 'r', imp.PY_SOURCE)
                    )
        
        assert simple_name in sys.modules
    
    def test_load_package(self):
        from instrumental.instrument import AnnotatorFactory
        from instrumental.metadata import MetadataGatheringVisitor
        from instrumental.monkey import load_module_factory
        from instrumental.pragmas import PragmaFinder
        from instrumental.recorder import ExecutionRecorder
        
        import instrumental.test.samples
        samples_directory = os.path.dirname(instrumental.test.samples.__file__)
        simple_name = 'instrumental.test.samples.package'
        simple_path = os.path.join(samples_directory, 'package')
        
        source = open(os.path.join(simple_path, '__init__.py'), "r").read()
        pragmas = PragmaFinder().find_pragmas(source)
        metadata = MetadataGatheringVisitor.analyze(self.config, 
                                                    simple_name, 
                                                    '__init__.py',
                                                    source, 
                                                    pragmas)
        
        recorder = ExecutionRecorder.get()
        recorder.add_metadata(metadata)
        visitor_factory = AnnotatorFactory(self.config, recorder)
        load_module = load_module_factory(['instrumental.test.samples.package'], 
                                          [],
                                          visitor_factory)
        
        load_module(simple_name,
                    None,
                    simple_path,
                    (None, None, imp.PKG_DIRECTORY)
                    )
        
        assert simple_name in sys.modules
        



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import os
import sys
import time

from xml.etree import ElementTree

import instrumental
from instrumental.constructs import BooleanDecision

class CoverageObject(object):
    
    @property
    def line_rate(self):
        statements = self.statements
        if statements:
            total_statements = len(statements)
            hit_statements = len([statement for statement in statements
                                  if statement])
            return float(hit_statements) / total_statements
        else:
            return 1.0
    
    @property
    def branch_rate(self):
        decisions = self.decisions
        if decisions:
            total_conditions = 0
            for decision in self.decisions:
                total_conditions += decision.number_of_conditions(self.options.report_conditions_with_literals)
            total_conditions_hit = 0
            for decision in self.decisions:
                total_conditions_hit += decision.number_of_conditions_hit()
            return float(total_conditions_hit) / total_conditions
        else:
            return 1.0
    
    @property
    def condition_rate(self):
        if self.constructs:
            total_conditions = sum(construct.number_of_conditions(self.options.report_conditions_with_literals)
                                   for construct in self.conditions)
            total_conditions_hit = sum(construct.number_of_conditions_hit()
                                       for construct in self.conditions)
            return float(total_conditions_hit) / total_conditions            
        else:
            return 1.0
    
    @property
    def decisions(self):
        return [construct for construct in self.constructs
                if construct.is_decision()]
    
    @property
    def conditions(self):
        return [construct for construct in self.constructs
                if not construct.is_decision()]
    
    def to_element(self):
        element = ElementTree.Element(self.tag)
        if hasattr(self, 'name'):
            element.attrib['name']  = self.name
        element.attrib['complexity'] = '0.0'
        element.attrib['line-rate'] = "%f" % self.line_rate
        element.attrib['branch-rate'] = "%f" % self.branch_rate
        element.attrib['condition-rate'] = "%f" % self.condition_rate
        return element

class Coverage(CoverageObject):
    tag = 'coverage'
    
    def __init__(self, working_directory, metadata, options):
        self._working_directory = working_directory
        self._metadata = metadata
        self.options = options
    
    @property
    def statements(self):
        _statements = []
        for modulename in self._metadata:
            _statements += self._metadata[modulename].lines.values()
        return _statements
    
    @property
    def constructs(self):
        return [construct
                for modulename in self._metadata
                for construct in self._metadata[modulename].constructs.values()]
                
    def _package_for_module(self, modulename):
        package_path = os.path.join(modulename.replace('.', os.path.sep),
                                    '__init__.py')
        if os.path.exists(package_path):
            return modulename
        else:
            return modulename.rsplit('.', 1)[0]
    
    def _path_for_module(self, modulename):
        package_path = os.path.join(modulename.replace('.', os.path.sep),
                                    '__init__.py')
        if os.path.exists(package_path):
            return package_path
        else:
            return modulename.replace('.', os.path.sep) + ".py"
        
    def to_element(self):
        coverage_element = super(Coverage, self).to_element()
        del coverage_element.attrib['complexity']
        coverage_element.attrib['timestamp'] = str(int(time.time()))
        coverage_element.version = instrumental.__version__
        
        comment = 'Generated by instrumental: http://bitbucket.org/desmaj/instrumental'
        coverage_element.append(ElementTree.Comment(comment))
        
        here = self._working_directory
        packages = {}
        for modulename in self._metadata:
            packagename = self._package_for_module(modulename)
            modules = packages.setdefault(packagename, [])
            modules.append((modulename, self._path_for_module(modulename)))
        
        packages_element = ElementTree.Element('packages')
        for packagename, modulenames in packages.items():
            package_coverage = PackageCoverage(packagename,
                                               packages[packagename],
                                               self._metadata,
                                               self.options)
            package_element = package_coverage.to_element()
            packages_element.append(package_element)
        
        coverage_element.append(packages_element)
        
        return coverage_element

class PackageCoverage(CoverageObject):
    tag = 'package'
    
    def __init__(self, name, modules, metadata, options):
        self.name = name
        self._modules = modules
        self._metadata = metadata
        self.options = options
    
    @property
    def statements(self):
        _statements = []
        for modulename, modulefile in self._modules:
            _statements += self._metadata[modulename].lines.values()
        return _statements
    
    @property
    def constructs(self):
        return [construct
                for modulename, modulefile in self._modules
                for construct in self._metadata[modulename].constructs.values()]
            
    def to_element(self):
        element = super(PackageCoverage, self).to_element()
        
        modules = ElementTree.Element('classes')
        element.append(modules)
        
        for modulename, modulefile in self._modules:
            module_constructs = (
                [construct for construct 
                 in self._metadata[modulename].constructs.values()])
            module_coverage = ModuleCoverage(modulename,
                                             modulefile,
                                             self._metadata[modulename],
                                             self.options)
            modules.append(module_coverage.to_element())
        
        return element
        
class ModuleCoverage(CoverageObject):
    tag = 'class'
    
    def __init__(self, name, filename, metadata, options):
        self.name = name
        self.filename = filename
        self._metadata = metadata
        self.options = options
    
    @property
    def statements(self):
        return self._metadata.lines.values()
    
    @property
    def constructs(self):
        return self._metadata.constructs.values()
    
    def to_element(self):
        element = super(ModuleCoverage, self).to_element()
        element.attrib['filename'] = self.filename
        element.append(ElementTree.Element('methods'))
        
        lines_element = ElementTree.Element('lines')
        for lineno, hit in sorted(self._metadata.lines.items()):
            line_element = ElementTree.Element('line')
            line_element.attrib['line'] = str(lineno)
            line_element.attrib['hits'] = str(int(hit))
            lines_element.append(line_element)
        element.append(lines_element)
        
        return element
    
class XMLCoverageReport(object):
    
    def __init__(self, working_directory, metadata, options):
        self._working_directory = working_directory
        self._metadata = metadata
        self.options = options
    
    def write(self, filename):
        tree = str(self)
        document = '\n'.join([
                '<?xml version="1.0" ?>',
                '<!DOCTYPE coverage ',
                "  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>",
                tree,
                ])
        file(filename, 'w').write(document)
    
    def __str__(self):
        return ElementTree.tostring(self.tree.getroot())
    
    @property
    def tree(self):
        coverage_element = Coverage(self._working_directory, self._metadata, self.options)
        coverage = coverage_element.to_element()
        
        _tree = ElementTree.ElementTree(coverage)
        return _tree
        



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from copy import copy
from copy import deepcopy
import inspect
import sys

from astkit import ast

from instrumental.pragmas import PragmaFinder

def __setup_recorder(): # pragma: no cover
    from instrumental.recorder import ExecutionRecorder
    _xxx_recorder_xxx_ = ExecutionRecorder.get()

def get_setup():
    source = inspect.getsource(__setup_recorder)
    mod = ast.parse(source)
    defn = mod.body[0]
    setup = defn.body[:]
    for stmt in setup:
        stmt.lineno -= 1
    return setup

class ExecutionRecorder(object):
    DEFAULT_TAG = 'X'
    
    @classmethod
    def reset(cls):
        cls._instance = None
    
    _instance = None
    @classmethod
    def get(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance
    
    def __init__(self):
        self.metadata = {}
        self.recording = False
        self.tag = None
        self._tagging = False
    
    def start(self):
        self.recording = True
    
    def stop(self):
        self.recording = False
    
    def add_metadata(self, metadata):
        self.metadata[metadata.modulename] = metadata
    
    @staticmethod
    def get_recorder_call():
        kall = ast.Call()
        kall.func = ast.Attribute(value=ast.Name(id="_xxx_recorder_xxx_",
                                                 ctx=ast.Load()),
                                  attr="record",
                                  ctx=ast.Load())
        kall.keywords = []
        return kall
    
    def record(self, arg, modulename, label, *args, **kwargs):
        if self.recording and label in self.metadata[modulename].constructs:
            kwargs['tag'] = (self.tag 
                             if self.tag is not None 
                             else self.DEFAULT_TAG)
            self.metadata[modulename].constructs[label].record(arg, 
                                                               *args, **kwargs)
        return arg
    
    def add_BoolOp(self, modulename, label, node, pragmas, parent):
        # Now wrap the individual values in recorder calls
        base_call = self.get_recorder_call()
        base_call.args = \
            [ast.Str(s=modulename, lineno=node.lineno, col_offset=node.col_offset),
             ast.Str(s=label, lineno=node.lineno, col_offset=node.col_offset)]
        for i, value in enumerate(node.values):
            recorder_call = deepcopy(base_call)
            recorder_call.args.insert(0, node.values[i])
            recorder_call.args.append(ast.copy_location(ast.Num(n=i), node.values[i]))
            node.values[i] = ast.copy_location(recorder_call, node.values[i])
        ast.fix_missing_locations(node)
        return node
    
    def add_test(self, modulename, label, node):
        base_call = ast.copy_location(self.get_recorder_call(),
                                      node)
        base_call.args = \
            [node,
             ast.Str(s=modulename, lineno=node.lineno, col_offset=node.col_offset),
             ast.Str(s=label, lineno=node.lineno, col_offset=node.col_offset)]
        ast.fix_missing_locations(base_call)
        return base_call
    
    def add_comparison(self, modulename, label, node):
        base_call = ast.copy_location(self.get_recorder_call(),
                                      node)
        base_call.args = \
            [node,
             ast.Str(s=modulename, lineno=node.lineno, col_offset=node.col_offset),
             ast.Str(s=label, lineno=node.lineno, col_offset=node.col_offset)]
        ast.fix_missing_locations(base_call)
        return base_call
    
    @staticmethod
    def get_statement_recorder_call(modulename, lineno):
        kall = ast.Call()
        kall.func = ast.Attribute(value=ast.Name(id="_xxx_recorder_xxx_",
                                                 ctx=ast.Load()),
                                  attr="record_statement",
                                  ctx=ast.Load())
        kall.args = [ast.Str(s=modulename),
                     ast.Num(n=lineno),
                     ]
        kall.keywords = []
        kall_stmt = ast.Expr(value=kall)
        return kall_stmt
    
    def record_statement(self, modulename, lineno):
        if self.recording and lineno in self.metadata[modulename].lines:
            self.metadata[modulename].lines[lineno] = True
    
    def add_statement(self, modulename, node):
        marker = self.get_statement_recorder_call(modulename, node.lineno)
        marker = ast.copy_location(marker, node)
        ast.fix_missing_locations(marker)
        return marker



from astkit import ast

#test_instrument_nodes.py
def test_Print(self):
    def test_module():
        print bar
    inst_module = self._instrument_module(test_module)
    self._assert_recorder_setup(inst_module)
    
    self._assert_record_statement(inst_module.body[2], 'test_module', 1)
    assert isinstance(inst_module.body[3], ast.Print)
    assert not inst_module.body[3].dest
    assert isinstance(inst_module.body[3].values[0], ast.Name)
    assert inst_module.body[3].values[0].id == 'bar'
    assert inst_module.body[3].nl



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
""" instrument.py - instruments ASTs representing Python programs
    
    We define instrument here to mean adding code that will have a side effect
    that we can measure so that we can determine when the code was executed.
    
    This is made difficult by the fact that python is so dynamic and that
    boolean operations have the properties that they do. The problems are that
    (a) the first non-True value will be returned from an and operation and the
    first non-False value will be returned from an or operation, (b) evaluation
    stops when the result of the operation has been determined.
    
"""
import sys

from astkit import ast
from astkit.render import SourceCodeRenderer

from instrumental import recorder
from instrumental.pragmas import PragmaNoCover

def force_location(tree, lineno, col_offset=0):
    for node in ast.walk(tree):
        if hasattr(node, 'lineno'):
            node.lineno = lineno
            node.col_offset = col_offset

def has_docstring(defn):
    return ast.get_docstring(defn) is not None

def has_future_import(module):
    if not module.body:
        return False
    return (isinstance(module.body[0], ast.ImportFrom)
            and module.body[0].module == '__future__')

class InstrumentedNodeFactory(object):
    
    def __init__(self, recorder):
        self._recorder = recorder
    
    def instrument_node(self, modulename, label, node, pragmas, parent):
        return self._recorder.add_BoolOp(modulename, label, node, pragmas, parent)
    
    def instrument_test(self, modulename, label, node):
        return self._recorder.add_test(modulename, label, node)
    
    def instrument_comparison(self, modulename, label, node):
        return self._recorder.add_comparison(modulename, label, node)
    
    def instrument_statement(self, modulename, node):
        return self._recorder.add_statement(modulename, node)

class AnnotatorFactory(object):
    
    def __init__(self, config, recorder):
        self.config = config
        self.recorder = recorder
    
    def create(self, modulename, module_source):
        return CoverageAnnotator(self.config, modulename, self.recorder)

class CoverageAnnotator(ast.NodeTransformer):
    
    def __init__(self, config, modulename, recorder):
        self.config = config
        self.modulename = modulename
        self.pragmas = recorder.metadata[modulename].pragmas
        self.node_factory = InstrumentedNodeFactory(recorder)
        self.modifiers = []
        self.expression_context = [None]
        self._found_labels = []
    
    def _next_label(self, lineno):
        i = 1
        while ('%s.%s' % (lineno, i)) in self._found_labels:
            i += 1
        label = '%s.%s' % (lineno, i)
        self._found_labels.append(label)
        return label
    
    def _has_pragma(self, lineno, pragma_klass):
        return any(isinstance(pragma, pragma_klass)
                   for pragma in self.pragmas[lineno])
    
    def visit_Module(self, module):
        recorder_setup = recorder.get_setup()
        docstring = None
        if has_docstring(module):
            docstring = module.body.pop(0)
        self.generic_visit(module)
        
        if has_future_import(module):
            future_import = module.body.pop(0)
            recorder_setup.insert(0, future_import)
        if docstring:
            recorder_setup.insert(0, docstring)
        for node in recorder_setup:
            force_location(node, 1)
        module.body = recorder_setup + module.body
        
        return module
    
    def visit_BoolOp(self, boolop):
        pragmas = self.pragmas.get(boolop.lineno, [])
        label = self._next_label(boolop.lineno)
        boolop = self.generic_visit(boolop)
        result =\
            self.node_factory.instrument_node(self.modulename, label, boolop, pragmas, None)
        return result
    
    def visit_Compare(self, compare):
        if self.config.instrument_comparisons:
            pragmas = self.pragmas.get(compare.lineno, [])
            label = self._next_label(compare.lineno)
            compare = self.generic_visit(compare)
            result = self.node_factory.instrument_comparison(self.modulename, label, compare)
        else:
            result = self.generic_visit(compare)
        return result
    
    def visit_IfExp(self, ifexp):
        label = self._next_label(ifexp.lineno)
        ifexp.test = self.node_factory.instrument_test(self.modulename, label, ifexp.test)
        result = self.generic_visit(ifexp)
        return result
    
    def _visit_stmt(self, node):
        if self._has_pragma(node.lineno, PragmaNoCover):
            self.modifiers.append(PragmaNoCover)
        
        if PragmaNoCover in self.modifiers:
            result = node
        else:
            marker = self.node_factory.instrument_statement(self.modulename, node)
            self.generic_visit(node)
            result = [marker, node]
        
        if self._has_pragma(node.lineno, PragmaNoCover):
            self.modifiers.pop(-1)
        return result
    
    def visit_Assert(self, assert_):
        if self.config.instrument_assertions:
            if isinstance(assert_.test, ast.BoolOp):
                label = self._next_label(assert_.lineno)
                assert_.test = self.node_factory.instrument_test(self.modulename, label, assert_.test)
            result = self._visit_stmt(assert_)
        else:
            result = assert_
        return result
    
    def visit_assignment(self, assign):
        if isinstance(assign.value, ast.BoolOp):
            label = self._next_label(assign.lineno)
            assign.value = self.node_factory.instrument_test(self.modulename, 
                                                             label, 
                                                             assign.value)
        return self._visit_stmt(assign)
    
    def visit_Assign(self, assign):
        return self.visit_assignment(assign)
    
    def visit_AugAssign(self, augassign):
        return self.visit_assignment(augassign)
    
    def visit_Break(self, break_):
        return self._visit_stmt(break_)
    
    def _visit_defn_with_docstring(self, defn):
        if PragmaNoCover in self.modifiers:
            result = defn
        else:
            # grab the docstring so that it isn't visited generically
            # docstrings don't seem to ever be 'executed', so we shouldn't
            # count them
            docstring = defn.body.pop(0)
            self.generic_visit(defn)
            
            # and put the docstring back
            defn.body = [docstring] + defn.body
            
            marker = self.node_factory.instrument_statement(self.modulename, defn)
            result = [marker, defn]
        return result
    
    def visit_ClassDef(self, defn):
        if not has_docstring(defn):
            visitor = self._visit_stmt
        else:
            visitor = self._visit_defn_with_docstring
            
        if self._has_pragma(defn.lineno, PragmaNoCover):
            self.modifiers.append(PragmaNoCover)
        
        if PragmaNoCover in self.modifiers:
            result = defn
        else:
            result = visitor(defn)
        
        if self._has_pragma(defn.lineno, PragmaNoCover):
            self.modifiers.pop(-1)
        
        return result
    
    def visit_Continue(self, continue_):
        return self._visit_stmt(continue_)
    
    def visit_Delete(self, delete):
        return self._visit_stmt(delete)
    
    def visit_Exec(self, exec_):
        return self._visit_stmt(exec_)
    
    def visit_Expr(self, expr):
        return self._visit_stmt(expr)
    
    def visit_ExceptHandler(self, excepthandler):
        if self._has_pragma(excepthandler.lineno, PragmaNoCover):
            self.modifiers.append(PragmaNoCover)
        
        if PragmaNoCover in self.modifiers:
            result = excepthandler
        else:
            result = self.generic_visit(excepthandler)
        
        if self._has_pragma(excepthandler.lineno, PragmaNoCover):
            self.modifiers.pop(-1)
        
        return result
    
    def visit_For(self, for_):
        if self._has_pragma(for_.lineno, PragmaNoCover):
            self.modifiers.append(PragmaNoCover)
        
        if PragmaNoCover in self.modifiers:
            result = for_
        else:
            self.generic_visit(for_)
            marker = self.node_factory.instrument_statement(self.modulename, for_)
            result = [marker, for_]
        
        if self._has_pragma(for_.lineno, PragmaNoCover):
            self.modifiers.pop(-1)
        
        return result
        
    def visit_FunctionDef(self, defn):
        if not has_docstring(defn):
            return self._visit_stmt(defn)
        
        if self._has_pragma(defn.lineno, PragmaNoCover):
            self.modifiers.append(PragmaNoCover)
        
        result = self._visit_defn_with_docstring(defn)
        
        if self._has_pragma(defn.lineno, PragmaNoCover):
            self.modifiers.pop(-1)
        
        return result
    
    def visit_Global(self, global_):
        return self._visit_stmt(global_)
    
    def visit_If(self, if_):
        if self._has_pragma(if_.lineno, PragmaNoCover):
            self.modifiers.append(PragmaNoCover)
        
        if PragmaNoCover in self.modifiers:
            result = if_
        else:
            label = self._next_label(if_.lineno)
            if_.test = self.node_factory.instrument_test(self.modulename, label, if_.test)
            if_ = self.generic_visit(if_)
            marker = self.node_factory.instrument_statement(self.modulename, if_)
            result = [marker, if_]
        
        if self._has_pragma(if_.lineno, PragmaNoCover):
            self.modifiers.pop(-1)
        return result
    
    def visit_Import(self, import_):
        return self._visit_stmt(import_)
    
    def visit_ImportFrom(self, import_):
        if self._has_pragma(import_.lineno, PragmaNoCover):
            self.modifiers.append(PragmaNoCover)
        
        self.generic_visit(import_)
        if PragmaNoCover in self.modifiers:
            result = import_
        else:
            marker = self.node_factory.instrument_statement(self.modulename, import_)
            if import_.module == '__future__':
                result = [import_, marker]
            else:
                result = [marker, import_]
        
        if self._has_pragma(import_.lineno, PragmaNoCover):
            self.modifiers.pop(-1)
        return result
    
    def visit_Pass(self, pass_):
        return self._visit_stmt(pass_)
    
    def visit_Print(self, print_):
        return self._visit_stmt(print_)
    
    def visit_Raise(self, raise_):
        return self._visit_stmt(raise_)
    
    def visit_Return(self, return_):
        return self._visit_stmt(return_)
    
    def visit_TryExcept(self, try_):
        return self._visit_stmt(try_)
    
    def visit_TryFinally(self, try_):
        return self._visit_stmt(try_)
    
    def visit_While(self, while_):
        if self._has_pragma(while_.lineno, PragmaNoCover):
            self.modifiers.append(PragmaNoCover)
        
        if PragmaNoCover in self.modifiers:
            result = while_
        else:
            label = self._next_label(while_.lineno)
            while_.test = self.node_factory.instrument_test(self.modulename, label, while_.test)
            self.generic_visit(while_)
            marker = self.node_factory.instrument_statement(self.modulename, while_)
            result = [marker, while_]
        
        if self._has_pragma(while_.lineno, PragmaNoCover):
            self.modifiers.pop(-1)
        return result

    def visit_With(self, with_):
        return self._visit_stmt(with_)



import contextlib
import datetime
import os
import shutil
import time

from instrumental import util
from instrumental.metadata import FileBackedMetadataCache
from instrumental.metadata import ModuleMetadata

@contextlib.contextmanager
def monkeypatch_now(when):
    old_now = util.now
    util.now = lambda: when
    yield
    util.now = old_now

class TestFileBackedMetadataCache(object):
    CACHE_PATH = '.instrumental.cache'
    
    def setup(self):
        self._dirty_paths = []# self.CACHE_PATH]
    
    def teardown(self):
        for path in self._dirty_paths:
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)
    
    def _make_one(self):
        return FileBackedMetadataCache()
    
    def _make_metadata(self, filepath):
        return ModuleMetadata(filepath,
                              filepath.replace(os.path.sep, '.') + '.py',
                              source = "",
                              pragmas = {})

    def _touch_file(self, filepath):
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        open(filepath, 'w').close()
    
    def test_new_cache(self):
        cache = self._make_one()
    
    def test_initialize(self):
        cache = self._make_one()
        cache.initialize()
        
        assert os.path.exists(self.CACHE_PATH)
    
    def test_item_presence_when_present_and_valid(self):
        cache = self._make_one()
        cache.initialize()
        
        filepath = os.path.join(os.getcwd(), 'path/to/module.py')
        self._touch_file(filepath)
        self._dirty_paths.append(filepath)
        
        meta = self._make_metadata(filepath)
        timestamp = datetime.datetime.now() + datetime.timedelta(seconds=1)
        with monkeypatch_now(timestamp):
            cache.store(filepath, meta)
        cached_metadata = cache.fetch(filepath)
        assert cached_metadata
    
    def test_item_absence_when_present_and_stale(self):
        cache = self._make_one()
        cache.initialize()
        
        filepath = os.path.join(os.getcwd(), 'path/to/module.py')
        self._touch_file(filepath)
        self._dirty_paths.append(filepath)
        
        meta = self._make_metadata(filepath)
        timestamp = datetime.datetime.now() + datetime.timedelta(seconds=-2)
        with monkeypatch_now(timestamp):
            cache.store(filepath, meta)
        self._touch_file(filepath)
        cached_metadata = cache.fetch(filepath)
        assert not cached_metadata



# 
# Copyright (C) 2012  Matthew J Desmarais

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import re

from astkit import ast

class Pragma(object):
    block = False
    
    def __init__(self, match):
        pass

    def apply(self, construct): # pragma: no cover
        return construct

class PragmaNoCover(Pragma):
    block = True
    
class PragmaNoCondition(Pragma):
    block = False
    
    NO_COND_TAG = 'P'
    
    def __init__(self, match):
        selector = match.group('selector')
        if selector:
            _, self.selector = selector.split('.')
        else:
            self.selector = '1'
        self.conditions = match.group('conditions').split(',')
        
    def apply(self, construct):
        for condition in construct.conditions:
            if construct.description(condition) in self.conditions:
                construct.conditions[condition].add(self.NO_COND_TAG)
        return construct

selector = r'(\[(?P<selector>\d*\.\d+)\])?'

no_cover_root = r'no\s?cover'
no_cover = no_cover_root + selector

no_cond_root = r'no\s+cond'
no_cond_condition = r'[TF\*](\s+[TF\*])*'
no_cond_conditions = r'\((?P<conditions>%s(,%s)*)\)' % (no_cond_condition,
                                                        no_cond_condition)
no_cond = no_cond_root + selector + no_cond_conditions

valid_pragmas = {
    no_cover: PragmaNoCover,
    no_cond: PragmaNoCondition,
    }

class PragmaApplier(ast.NodeVisitor):
    """ Expand source-level line-pragmas relations to node-level relations
    
        This visitor starts with a mapping of source lines to the pragmas that
        should be applied to those source lines. Calling `apply` will start
        a traversal of the an AST during which source line and node
        dependencies will be used to determine to which lines the defined
        pragmas should be applied.
        
        This visitor should take into account that statements can be spread
        over multiple source lines and should produce a mapping with the
        pragmas for any given source line being applied to the line of the
        statement that contains the initial source line. (Think argument
        lists that cover more than one line).
        
        Example:
        1: if condition:
        2:     pass
        3: else: # pragma: no cover
        4:     do_something()
        
        In this example, the intent is that the 'no cover' pragma should be
        applied to the statements on the 'else' path of the 'if' statement.
        There is no "Else" node type defined by the ast module, however, so
        we need to do some math. We'll figure out that the pragma is pointing
        at the 'else' branch of the 'if' statement and augment our pragma
        mapping to say that the original pragma should be applied to all of
        the statements in the body of the 'if' statement's "orelse" attribute.
        
    """
    
    def __init__(self, pragmas, source):
        self._base_pragmas = pragmas.copy()
        self._source = source
        self._statement_end_linenos = []
    
    def apply(self):
        self._pragmas = self._base_pragmas.copy()
        
        node = ast.parse(self._source)
        
        # fix the else cases
        self.visit(node)
        
        return self._pragmas
    
    def _block_pragmas(self, pragmas):
        return [pragma for pragma in pragmas
                if pragma.block]
    
    def _block_pragmas_for_range(self, start, end):
        line_range = range(start, end)
        pragma_lines = [lineno for lineno in self._pragmas
                        if self._block_pragmas(self._pragmas[lineno])]
        pragmas = set()
        for lineno in line_range:
            if lineno in pragma_lines:
                pragmas.update(self._block_pragmas(self._pragmas[lineno]))
        return pragmas
    
    def _add_pragmas_to_body(self, pragmas, body):
        for stmt in body:
            self._pragmas[stmt.lineno].update(pragmas)
    
    def _visit_node_with_body_and_else(self, node):
        self.generic_visit(node)
        
        stmt_pragmas = self._block_pragmas_for_range(node.lineno, 
                                                     node.body[0].lineno)
        if stmt_pragmas:
            for stmt in node.body:
                self._pragmas.setdefault(stmt.lineno, set())\
                    .update(stmt_pragmas)
        if node.orelse:
            else_pragmas = self._block_pragmas_for_range(node.body[-1].lineno+1, 
                                                         node.orelse[0].lineno)
            else_pragmas.update(stmt_pragmas)
            for stmt in node.orelse:
                self._pragmas.setdefault(stmt.lineno, set())\
                    .update(else_pragmas)
    visit_For = visit_If = _visit_node_with_body_and_else
    visit_While = _visit_node_with_body_and_else
    
    def _visit_node_with_body(self, node):
        self.generic_visit(node)
        
        stmt_pragmas = self._block_pragmas_for_range(node.lineno, 
                                                     node.body[0].lineno)
        if stmt_pragmas:
            self._add_pragmas_to_body(stmt_pragmas, node.body)
    visit_FunctionDef = visit_ClassDef = visit_With = _visit_node_with_body
    visit_ExceptHandler = visit_TryFinally = _visit_node_with_body
    
    def visit_TryExcept(self, node):
        self.generic_visit(node)
        
        stmt_pragmas = self._block_pragmas_for_range(node.lineno, 
                                                     node.body[0].lineno)
        if stmt_pragmas:
            for stmt in node.body:
                self._pragmas.setdefault(stmt.lineno, set())\
                    .update(stmt_pragmas)
        if node.orelse:
            else_pragmas = self._block_pragmas_for_range(node.handlers[-1].body[-1].lineno+1, 
                                                         node.orelse[0].lineno)
            else_pragmas.update(stmt_pragmas)
            for stmt in node.orelse:
                self._pragmas.setdefault(stmt.lineno, set())\
                    .update(else_pragmas)
    
class PragmaFinder(object):
    
    def __init__(self):
        pass
    
    def find_pragmas(self, source):
        pragmas = {}
        lines = source.splitlines()
        for lineno in range(1, len(lines)+1):
            pragmas[lineno] = set()
        for lineno, line in enumerate(lines):
            lineno += 1
            pragma_match = re.search(r'#\s*pragma:(.+)$', line)
            if pragma_match:
                pragma_text = pragma_match.group(1)
                for pragma_label, pragma in valid_pragmas.items():
                    pragma_match = re.search(pragma_label, pragma_text)
                    if pragma_match:
                        selected_lineno = lineno
                        selector = pragma_match.group('selector')
                        if selector:
                            selector_lineno, _ = selector.split('.')
                            if selector_lineno:
                                selected_lineno = int(selector_lineno)
                        pragmas[selected_lineno].add(pragma(pragma_match))
        
        applier = PragmaApplier(pragmas, source)
        pragmas = applier.apply()
        
        return pragmas
